'''
Generate a simple Hydra io network for specified io channels and io groups.
In this case, a 'simple' network applies the following constraints on the network:

     - All configuration packets originate at one point (head) and all data packets
     are returned via the same point
     - Hydra nodes are arranged in a regular grid, i.e. left 1, down 1 == down 1, left 1
     - Each uart is configured s.t. mosi[i] == miso_ds[i] or miso_us[i]
     - Neighboring nodes are configured s.t. miso_us[i] == mosi[j] with j being the connected
     port to i (e.g. 0<->2 and 1<->3)

'''

import numpy as np
import networkx as nx
from bidict import bidict
import random
from copy import deepcopy

default_miso_us_uart_map = [3,0,1,2]
default_miso_ds_uart_map = [1,2,3,0]
default_mosi_uart_map = [2,3,0,1]
uart_us_table = bidict({
    (0,1): 3,
    (1,0): 2,
    (-1,0): 0,
    (0,-1): 1
    })
uart_ds_table = bidict({
    (0,-1): 1,
    (-1,0): 0,
    (1,0): 2,
    (0,1): 3
    })

def edge_dir(edge):
    dx = edge[1][0] - edge[0][0]
    dy = edge[1][1] - edge[0][1]
    return (dx, dy)

def chip_id_simple(g, position):
    '''
    Generates an easy-to-read chip id based on the position (excluding 0, 1, and 255)
    Chip ids increase by increments of 1 at each node position in y and 10 at each
    node position in x (works for max of 8x8 grid)

    '''
    return (position[0]*10 + position[1]) % 253 + 2

def chip_id_position(g, position):
    '''
    Generates a chip id based on the position excluding ids 0, 1, and 255
    Chip id increase by increments of 1 at each node position in y dimension
    Chip ids are only guaranteed to be unique if less than 253 positions in
    rectangle bounding all nodes

    '''
    max_pos = max(g.nodes())
    min_pos = min(g.nodes())
    id_factor = max_pos[1] - min_pos[1] + 1
    return (position[0]*id_factor + position[1]) % 253 + 2

def generate_chip_id(g, method='chip_id_simple'):
    '''
    Generates a ``chip_id`` field on each node based a dedicated scheme

    '''
    for node in g.nodes():
        if 'chip_id' in g.nodes[node]:
            continue
        g.nodes[node]['chip_id'] = globals()[method](g, node)

def generate_digraphs_from_upstream(g, root, ext):
    '''
    Generate g_us, g_ds, g_mosi assuming g is a directed graph of the upstream
    network with root node ``root`` and an external node ``ext``

    :returns: tuple of g_us, g_ds, g_mosi

    '''
    # generate upstream miso
    g_us = nx.DiGraph()
    g_us.add_nodes_from(g.nodes())
    g_us.add_edges_from([edge for edge in g.edges()])
    g_us.nodes[root]['root'] = True
    g_us.nodes[ext]['chip_id'] = 'ext'

    # generate downstream miso
    g_ds = nx.DiGraph()
    g_ds.add_nodes_from(g_us.nodes())
    g_ds.add_edges_from([(edge[::-1]) for edge in g_us.edges()])
    g_ds.nodes[root]['root'] = True
    g_ds.nodes[ext]['chip_id'] = 'ext'

    # generate mosi
    g_mosi = nx.DiGraph()
    g_mosi.add_nodes_from(g_us.nodes())
    g_mosi.add_edges_from(g_us.edges())
    g_mosi.add_edges_from(g_ds.edges())
    g_mosi.nodes[root]['root'] = True
    g_mosi.nodes[ext]['chip_id'] = 'ext'

    return g_us, g_ds, g_mosi

def generate_2d_grid_digraph(x_slots, y_slots, *args):
    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph)
    for node in args:
        g.add_node(node)
    for x in range(-1,2):
        for y in range(-1,2):
            if abs(x) + abs(y) == 1:
                for node in args:
                    if (node[0]+x,node[1]+y) in g.nodes():
                        g.add_edge(node,(node[0]+x,node[1]+y))
                        g.add_edge((node[0]+x,node[1]+y),node)
    return g

def generate_hydra_simple_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345):
    '''
    Generates a simple hydra io networks using the simple algorithm. Roughly,
    this algorithm generates a simple hydra io network based on an upstream
    network. The upstream network is found by this algorithm:

        1. create network with all connections to neighbors

        2. starting at root node, prune all incoming connections that also have an outgoing connection

        3. iteratively prune connection with the largest n_children[head] + n_children[tail]

    There is not a unique solution and so a random seed is used to generate the network

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))

    # work forward from root node pruning out links that contribute to nodes
    # with largest fifo load
    nodes = [node for node in g.nodes() if g.in_degree(node) > 1]
    while nodes:
        edges = list(g.in_edges(nodes))
        random.shuffle(edges)

        metric = [len(nx.descendants(g, edge[1])) + len(nx.descendants(g, edge[0])) for edge in edges]
        g.remove_edge(*edges[np.argmax(metric)])
        nodes = [node for node in g.nodes() if g.in_degree(node) > 1]

    return generate_digraphs_from_upstream(g, root, ext)

def L(i, d_in, n_us):
    '''
    Helper function for calculating fifo load, representing the increase in the
    fifo value during readout period ``i``. Maximum FIFO value can be found
    by summing from ``i=-1`` to ``i=d_in-1``, inclusive.

    :param i: iterator index

    :param d_in: in degree of node in downstream network

    :param n_us: sorted list of number of downstream parents of node (0 == least upstream children) of each connection in downstream network

    :returns: ``int``

    '''
    if i == -1:
        return 1
    if i == 0:
        return d_in - 1
    if i == 1:
        return (d_in - i) * (n_us[0] - 1)
    return (d_in - i) * (n_us[i-1] - n_us[i-2])

def fifo_load(g, node):
    '''
    Calculates the "FIFO load" for a given node assuming g represents the
    upstream network, and the downstream network is equivalent to the inverse
    of the upstream network.

    The fifo load is a metric representing the relative maximum fifo value reached
    by a single chip assuming all chips trigger at the same time.

    :param g: directed graph representing upstream network

    :param node: node within ``g`` to calculate fifo load

    :returns: ``int`` of fifo load

    '''
    in_links = g.out_edges(node)
    n_inputs = len(in_links)
    n_outputs = 1
    n_upstream_nodes = sorted([len(nx.descendants(g, link[1]))+1 for link in in_links])
    return sum([L(i, n_inputs, n_upstream_nodes) for i in range(-1,n_inputs)])

def network_fifo_load_score(g):
    '''
    Creates the network fifo load score as min(fifo load) + mean(fifo load).
    Modifies the ``'load'`` attribute of each node in order to save on
    computation time. When modifying the network connection (thus the fifo load),
    the ``'load'`` attributes should be cleared (see methods ``XX_edge_and_delete_fifo_load``).

    :param g: directed graph representing upstream network

    :returns: ``float``

    '''
    fifo_loads = []
    n = 0
    sum = 0
    max = 0
    for node in g.nodes():
        if not 'load' in g.nodes[node]:
            g.nodes[node]['load'] = fifo_load(g, node)
        if g.nodes[node]['load'] > max:
            max = g.nodes[node]['load']
        n += 1
        sum += g.nodes[node]['load']
    return max + sum / n

def add_edge_and_delete_fifo_load(g, edge):
    '''
    Creates edge in network and deletes the ``'load'`` attr from all nodes upstream
    of connection.

    :param g: directed graph of upstream network

    :param edge: edge to add to network

    '''
    g.add_edge(*edge)
    nodes = [edge[0]]
    while nodes:
        for node in nodes:
            if 'load' in g.nodes[node]:
                del g.nodes[node]['load']
        nodes = list(set(edge[0] for edge in g.in_edges(nodes)))

def remove_edge_and_delete_fifo_load(g, edge):
    '''
    Removes edge from network and deletes the ``'load'`` attr from all nodes
    upstream of connection.

    :param g: directed graph of upstream network

    :param edge: edge to remove from network

    '''
    g.remove_edge(*edge)
    nodes = [edge[0]]
    while nodes:
        for node in nodes:
            if 'load' in g.nodes[node]:
                del g.nodes[node]['load']
        nodes = list(set(edge[0] for edge in g.in_edges(nodes)))

def random_prune(g):
    '''
    Simple pruning algorithm:

        1. Find all edges going into nodes with in degree > 1

        2. Randomly remove (as long as all nodes maintain in degree > 1)

    :param g: directed graph representing upstream network

    :returns: ``g``

    '''
    edges_to_prune = [edge for edge in g.in_edges() if g.in_degree(edge[1]) > 1]
    random.shuffle(edges_to_prune)
    for edge in edges_to_prune:
        if g.in_degree(edge[1]) > 1:
            g.remove_edge(*edge)
    return g

def check_peturbations(g, edges, size, n, peturbation_delta=1e-6, pruning_algorithm=random_prune, scoring_algorithm=network_fifo_load_score):
    '''
    Check to see if a perturbations of the graph linkage improves (minimizes) performance

    :param g: directed graph of upstream network

    :param edges: edges to potentially add back into network (edges that already exist in network are ignored)

    :param size: number of edges to add back into network

    :param n: number of random perturbations to apply

    :param peturbation_delta: fractional change required to continue iteration

    :param pruning_algorithm: dedicated pruning algorithm to use (should take a single upstream directed graph as an argument and return a upstream single directed graph)

    :param scoring_algorithm: algorithm used to determine performance, this is the function we are minimizing (should take a single upstream directed graph as an argument and return a number)

    :returns: best performing graph found during iterations

    '''
    return_g = g
    while True:
        print('trying {} peturbations of scale {}...'.format(n, size))
        base_performance = network_fifo_load_score(return_g) # performance of iteration graph
        current_performance = base_performance # best performance of peturbation graphs
        new_edges = [edge for edge in edges if edge not in return_g.edges()]

        for i in range(n):
            random.shuffle(new_edges)
            test_g = deepcopy(return_g)
            for edge in new_edges[:min(size,len(new_edges))]:
                add_edge_and_delete_fifo_load(test_g, edge)
            test_g = pruning_algorithm(test_g)
            test_performance = network_fifo_load_score(test_g)
            if test_performance < current_performance:
                print('improved performance to {}'.format(test_performance))
                return_g = test_g
                current_performance = test_performance
        performance_delta = abs(current_performance - base_performance)/base_performance
        if performance_delta < peturbation_delta:
            break
        base_performance = current_performance
    return return_g

def generate_hydra_min_fifo_load_beam_search_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=1e-6, beam_width=3):
    '''
    Generates a simple hydra io networks through the following algorithm:

        1. create network with all connections between neighbors

        2. starting from root node, remove all incoming connections that have an outgoing connection

        3. iteratively prune remaing connections following:

            1. calculate a weight for each redundant connection based on the fifo load score of the network with and without the connection

            2. generate a list of potential graphs (with N-1 edges) that minimize the fifo load score

            3. keep ``beam_width`` of the best performing potential graphs for next iteration

            4. iterate until all graphs have no more redundant connections (choosing the most performant)

        4. perturb the solution N/4 times by

            1. randomly add 25% of the connections pruned in step (3.)

            2. reprune

            3. keep most performant

        5. continue to perturb until the fractional improvement drops below ``peturbation_delta``

    There is not a unique solution and so a random seed is used to generate the
    network.

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :param peturbation_delta: continue to try peturbations if network score continues to change by this much (fractional, if None, don't check peturbations)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)
    
    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    shortest_paths = deepcopy(list(g.edges()))

    def prune_networks(stack, beam_width=beam_width):
        '''
        Prunes redundant incoming network connections of an upstream directed
        graph using a beam search-like algorithm.

        :param stack: a directed graph or a list of directed graphs representing the upstream network

        :param beam_width: length of stack used during search

        :returns: directed graph with redundant incoming network connections pruned

        '''
        if not isinstance(stack, list):
            stack = [stack]
        stack.sort(key=lambda x: network_fifo_load_score(x))
        while len(stack) > beam_width:
            del stack[-1]
        completed_flag = True
        next_stack = []
        for g in stack:
            edges = [edge for edge in g.in_edges() if g.in_degree(edge[1]) > 1]
            if not edges:
                next_stack += [g]
            else:
                for idx,edge in enumerate(edges):
                    remove_edge_and_delete_fifo_load(g, edge)
                    score = network_fifo_load_score(g)
                    add_edge_and_delete_fifo_load(g, edge)
                    g.edges[edge]['weight'] = score
                min_weight = min([g.edges[edge]['weight'] for edge in edges])
                edges_to_prune = [edge for edge in edges if g.edges[edge]['weight'] == min_weight]
                if not edges_to_prune:
                    next_stack += [g]
                elif len(edges_to_prune) == 1:
                    remove_edge_and_delete_fifo_load(g, edges_to_prune[0])
                    completed_flag = False
                    next_stack += [g]
                else:
                    random.shuffle(edges_to_prune)
                    for edge in edges_to_prune:
                        next_stack += [deepcopy(g)]
                        remove_edge_and_delete_fifo_load(next_stack[-1], edge)
                        completed_flag = False
        if completed_flag:
            next_stack.sort(key=lambda x: network_fifo_load_score(x))
            return next_stack[0]
        return prune_networks(next_stack, beam_width=beam_width)

    g = prune_networks(g)

    if not peturbation_delta is None:
        size = int(max(len(shortest_paths)/4,1))
        n = int(max(len(shortest_paths)/4,1))
        g = check_peturbations(g, shortest_paths, size, n, peturbation_delta=peturbation_delta, pruning_algorithm=prune_networks, scoring_algorithm=network_fifo_load_score)

    return generate_digraphs_from_upstream(g, root, ext)

# Handle for the default configuration (use beam search and peturbations)
generate_hydra_min_fifo_load_digraphs = generate_hydra_min_fifo_load_beam_search_digraphs

def generate_hydra_min_fifo_load_base_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=None, beam_width=1):
    '''
    Handle for the ``generate_hydra_min_fifo_load_beam_search_digraphs``
    algorithm with diffent default arguments.

    '''
    return generate_hydra_min_fifo_load_beam_search_digraphs(x_slots, y_slots, root, ext, avoid=avoid, seed=seed, peturbation_delta=peturbation_delta, beam_width=beam_width)

def greedy_tree(g, root):
    '''
    Simple tree creation network for a grid-like set of nodes:

        1. Starting a root node, make all connections possible to neighbors

        2. Iterate on children created

    Insures shortest path to each node and meets simple hydra network requirements

    :param g: directed graph to add upstream connections

    :param root: root node used to start algorithm

    :returns: ``g``

    '''
    nodes = [root]
    while nodes:
        potential_links = []
        for node in nodes:
            children = [child for child in g.nodes() \
            if ((abs(node[0]-child[0]) == 1 and abs(node[1]-child[1]) == 0) \
            or (abs(node[0]-child[0]) == 0 and abs(node[1]-child[1]) == 1)) \
            and not (child == root or child == root)]
            potential_links += [(node,child) for child in children]
        random.shuffle(potential_links)
        for link in potential_links:
            if not nx.has_path(g,root,link[1]):
                g.add_edge(*link)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    return g

def generate_hydra_fifo_product_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345, peturbation_delta=1e-6):
    '''
    Generates a simple hydra io networks through the following algorithm:

        1. create network with all connections between neighbors

        2. starting from root node, remove all incoming connections that have an outgoing connection

        3. create maximally spanning arborescence using fifo_load(head) * fifo_load(tail) as the weight

        4. perturb the solution N times by

            1. randomly adding 25% of the connections pruned in step (3.)

            2. reprune

            3. keep most performant

        5. continue to perturb until the fractional improvement drops below ``peturbation_delta``

    There is not a unique solution and so a random seed is used to generate the
    network.

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :param peturbation_delta: continue to try peturbations if network score continues to change by this much (fractional, if None, don't check peturbations)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''

    random.seed(seed)

    g = generate_2d_grid_digraph(x_slots, y_slots, root, ext)
    #    g = nx.grid_2d_graph(x_slots, y_slots, create_using=nx.DiGraph) # base graph for upstream

    if avoid:
        g.remove_nodes_from(avoid)

    # create root
    g.remove_edges_from(list(g.in_edges(root)))

    # remove in edges from each successor
    nodes = list(g.successors(root))
    while nodes:
        random.shuffle(nodes)
        edges_to_remove = [edge for edge in g.in_edges(nodes) \
            if edge[::-1] in g.out_edges(nodes) \
            and g.in_degree(edge[1]) > 1]
        g.remove_edges_from(edges_to_remove)
        nodes = list(set([edge[1] for edge in g.out_edges(nodes)]))
    shortest_paths = deepcopy(list(g.edges()))

    # calculate fifo product
    network_fifo_load_score(g)
    for edge in g.edges():
        g.edges[edge]['weight'] = g.nodes[edge[0]]['load'] * g.nodes[edge[1]]['load']

    g = nx.maximum_spanning_arborescence(g)

    if not peturbation_delta is None:
        size = int(max(len(shortest_paths)/4,1))
        n = len(shortest_paths)
        g = check_peturbations(g, shortest_paths, size, n, peturbation_delta=peturbation_delta, pruning_algorithm=random_prune, scoring_algorithm=network_fifo_load_score)

    return generate_digraphs_from_upstream(g, root, ext)


def generate_hydra_greedy_tree_digraphs(x_slots, y_slots, root, ext, avoid=None, seed=12345):
    '''
    Generates simple hydra io networks via the simple algorithm:

        1. starting a root node, make all outgoing connections to neighbors

        2. iterating on resulting children, make all outgoing connections to neighbors (as long as the in degree of the neighbor is < 1)

    :param x_slots: maximum number of x positions for nodes

    :param y_slots: maximum number of y positions for nodes

    :param root: ``(x,y)`` position for root node, ``0 <= x < x_slots``, ``0 <= y < y_slots``

    :param ext: position of the node representing the external system

    :param avoid: ``list`` of ``(x,y)`` nodes to exclude from network

    :param seed: random seed for configuration (insures repeatabilty)

    :returns: 3 networkx ``DiGraph``s indicating the upstream miso network, the downstream miso network, and the mosi network

    '''
    random.seed(seed)

    g = nx.DiGraph()
    for x in range(x_slots):
        for y in range(y_slots):
            g.add_node((x,y))
    g.add_node(ext)
    g.add_node(root)

    if avoid:
        g.remove_nodes_from(avoid)

    g = greedy_tree(g, root)

    return generate_digraphs_from_upstream(g, root, ext)


def generate_network_config(network_graphs, io_group, io_channel, name=None, previous=None):
    '''
    Converts a hydra network configuration graph into a json-encodable dict
    conforming to the Controller network configuration specification

    :param network_graphs: 3-tuple of networkx graph objects representing the miso_us, miso_ds, and mosi networks. Each node must have ``'chip_id'`` fields

    :param io_group: The io group for this hydra network

    :param io_channel: This io channel for this hydra network

    :param name: Name of configuration, default is ``max(nodes)``

    :param previous: A dict conforming to the Controller network configuration specification to add/overwrite

    :returns: json-ready ``dict`` conforming to Controller network configuration specification

    '''
    g = network_graphs[0] # upstream (convienence handle)
    network_config = {
        '_config_type': 'controller',
        'name': '{}'.format(max(g.nodes())),
        'asic_version': 2,
        'layout': None,
        'network': {}
        }
    if previous:
        network_config = previous
    if name:
        network_config['name'] = name
    io_group, io_channel = str(io_group), str(io_channel)
    if not io_group in network_config['network']:
        network_config['network'][io_group] = dict()
    if not io_channel in network_config['network'][io_group]:
        network_config['network'][io_group][io_channel] = dict()
    network_config['network'][io_group][io_channel]['nodes'] = []

    if not 'miso_us_uart_map' in network_config['network']:
        network_config['network']['miso_us_uart_map'] = default_miso_us_uart_map
    if not 'miso_ds_uart_map' in network_config['network']:
        network_config['network']['miso_ds_uart_map'] = default_miso_ds_uart_map
    if not 'mosi_uart_map' in network_config['network']:
        network_config['network']['mosi_uart_map'] = default_mosi_uart_map

    subnetwork = network_config['network'][io_group][io_channel]
    if not default_miso_us_uart_map == network_config['network']['miso_us_uart_map']:
        subnetwork['miso_us_uart_map'] = default_miso_us_uart_map
    if not default_miso_ds_uart_map == network_config['network']['miso_ds_uart_map']:
        subnetwork['miso_ds_uart_map'] = default_miso_ds_uart_map
    if not default_mosi_uart_map == network_config['network']['mosi_uart_map']:
        subnetwork['mosi_uart_map'] = default_mosi_uart_map

    next_nodes = [node for node in g.nodes() if 'root' in g.nodes[node]]
    while next_nodes:
        for node in next_nodes:
            subnetwork['nodes'] += [{
                'chip_id': g.nodes[node]['chip_id'],
                'miso_us': [None]*4
                }]
            chip_config = subnetwork['nodes'][-1]
            for child in g.successors(node):
                child_uart = uart_us_table[edge_dir((node,child))]
                chip_config['miso_us'][default_miso_us_uart_map.index(child_uart)] = g.nodes[child]['chip_id']
            if not any(chip_config['miso_us']):
                del chip_config['miso_us']
            if 'root' in g.nodes[node]:
                chip_config['root'] = True
        next_nodes = [edge[1] for edge in g.out_edges(next_nodes)]
    return network_config

def plot(g, name='default', labels='chip_id', figsize=(6, 6), interactive=True):
    '''
    Draws a network configuration graph (using matplotlib)

    :param g: A networkx graph object representing the configuration, each node must be a 2-``tuple`` with a ``'chip_id'`` field

    :param name: Figure name (if `'default'` will clear plot before redrawing)

    :param labels: Attribute to label nodes by, if None, don't label nodes

    :param figsize: Size of figure to draw

    :param interactive: turn on pyplot's interactive mode

    '''
    import matplotlib.pyplot as plt
    if interactive:
        plt.ion()
    plt.figure(name, figsize=figsize)
    if name == 'figure':
        plt.cla()
    if labels:
        nx.draw_networkx(g, dict([(node,node) for node in g.nodes]), labels=dict([(node,g.nodes[node][labels]) for node in g.nodes]))
    else:
        nx.draw_networkx(g, dict([(node,node) for node in g.nodes]), labels=dict([(node,'') for node in g.nodes]))
    plt.tight_layout()
    if not interactive:
        plt.savefig(name+'.png')

if __name__ == '__main__':
    import argparse
    import os
    import json

    parser = argparse.ArgumentParser(usage=__doc__)
    parser.add_argument('--outfile', '-o', type=str, required=False, default=None,
        help='''
        output filename (optional, default=<name>.json)
        if file exists, io groups / io channels will be added to existing file
        ''')
    parser.add_argument('--name', '-n', type=str, required=True, help='''
        configuration name (required)
        ''')
    parser.add_argument('--seed', '-s', type=int, required=False, default=12345, help='''
        seed for random shuffles
        ''')
    parser.add_argument('--x_slots', '-x', type=int, required=True, help='''
        max number of slots for network nodes (in x dim)
        ''')
    parser.add_argument('--y_slots', '-y', type=int, required=True, help='''
        max number of slots for network nodes (in y dim)
        ''')
    parser.add_argument('--root', '-r', type=json.loads, required=False, default=None, help='''
        position for root node as list '[x,y]', (optional, default=<ext node>)
        ''')
    parser.add_argument('--ext', '-e', type=json.loads, required=True, help='''
        position for ext node as list '[x,y]', (optional, default=<ext node>)
        ''')
    parser.add_argument('--avoid', type=json.loads, required=False, default=list(), help='''
        positions of nodes to avoid, specified as list of x, y positions '[[x,y],[x,y]]'
        ''')
    parser.add_argument('--chip_id', type=str, required=False, default='chip_id_simple',
        help='''sets method for generating and assigning chip ids, options are
        'chip_id_simple' and 'chip_id_position'
        ''')
    parser.add_argument('--algorithm', type=str, required=False, default='min_fifo_load',
        help='''sets method for generating network links, options are 'min_fifo_load', 'min_fifo_load_base', 'min_fifo_load_beam_search', 'fifo_product', 'greedy_tree', and 'simple'
        ''')
    parser.add_argument('--io_group', type=int, required=False, default=1,
        help='''sets io group'
        ''')
    parser.add_argument('--io_channel', type=int, required=False, default=1,
        help='''sets io channel'
        ''')
    parser.add_argument('--plot', action='store_true', help='''
        flag to plot network instead of generating file
        ''')

    args = parser.parse_args()
    outfile = args.outfile
    if not outfile:
        outfile = args.name + '.json'
    avoid = [tuple(pos) for pos in args.avoid]
    ext = tuple(args.ext)
    root = tuple(args.root) if args.root is not None else ext
    g_us, g_ds, g_mosi = globals()['generate_hydra_' + args.algorithm + '_digraphs'](args.x_slots,
        args.y_slots, root, ext, avoid=avoid, seed=args.seed)

    generate_chip_id(g_us, method=args.chip_id) # add chip ids to each node
    generate_chip_id(g_ds, method=args.chip_id) # add chip ids to each node
    generate_chip_id(g_mosi, method=args.chip_id) # add chip ids to each node

    fifo_loads = [fifo_load(g_us, node) for node in g_us.nodes()]
    n_children = [len(nx.descendants(g_us,node)) for node in g_us.nodes()]
    print('nodes: ', len(fifo_loads))
    print('mean fifo load: ', np.mean(fifo_loads))
    print('max fifo load: ', max(fifo_loads))

    if args.plot:
        plot(g_us, name='MISO upstream')
        plot(g_ds, name='MISO downstream')
        plot(g_mosi, name='MOSI')
        print('Press enter to quit')
        input()
        exit()

    else:
        plot(g_us, name=outfile.strip('.json'), interactive=False)
        
        previous_config = None
        if os.path.exists(outfile):
            print('loading existing file',outfile)
            with open(outfile,'r') as f:
                previous_config = json.load(f)
        new_config = generate_network_config((g_us, g_ds, g_mosi), args.io_group, args.io_channel, name=args.name, previous=previous_config)
        with open(outfile,'w') as f:
            json.dump(new_config, f, indent=4)



'''
A module to assist loading serial data logs for debugging
'''
from __future__ import absolute_import
import os

from .dataformatter import DataFormatter

class DataLoader(object):
    '''Simple low-level serial data log file loader'''
    def __init__(self, filename=None):
        '''Constructor'''
        self.filename = filename
        self.formatter = None
        self.bytes_read = 0
        self.file = None
        self._initialize()
        
    def _initialize(self):
        '''Initialize log for reading'''
        if not os.path.isfile(self.filename):
            raise ValueError('File does not exist: "%s"' % self.filename)
        # Set preliminary formatter
        self.formatter = DataFormatter
        # Open file
        self.open()
        # First block: file header, with data version
        byte_chunk = self.read_chunk()
        header_type = self.formatter.header_type(byte_chunk)
        if header_type != 'file':
            raise ValueError('First block in "%s" is not a file header!' %
                             self.filename)
        (major_version, minor_version) = self.formatter.data_version(
            byte_chunk)
        # Set formatter to correct version
        self.formatter = self.formatter.get_formatter_by_version(
            major_version,minor_version)
        self.close()
        return

    def read_chunk(self, n_chunks=1):
        '''Read one chunk from file stream'''
        if not self.is_open():
            # Open file if not yet opened
            self.open()
        raw_chunk = self.file.read(self.formatter.chunk_size * n_chunks)
        if len(raw_chunk) == 0:
            print('%%%%%%%%%%%% Reached end of log %%%%%%%%%%%%')
            return None
        if len(raw_chunk) < self.formatter.chunk_size:
            print('Warning: file is incorrectly truncated')
            return None
        data_chunk = bytearray(raw_chunk)
        self.bytes_read += len(data_chunk)
        return data_chunk

    def next_block_bytes(self):
        '''Load and return next serial data block as bytes'''
        head_chunk = self.read_chunk()
        if head_chunk is None:
            return None
        head_size = self.formatter.header_size(head_chunk)
        # If header size greater than 1, read rest of header
        if head_size > 1:
            head_chunk += self.read_chunk(n_chunks=(head_size-1))
        # Determine block size, and read rest of block if needed
        block_size = self.formatter.block_size(head_chunk)
        block_chunk = head_chunk[:]
        if block_size > head_size:
            block_chunk += self.read_chunk(block_size - head_size)
        if self.formatter.data_continued(block_chunk):
            # Catch continued blocks, and append data
            block_chunk += self.next_block_bytes()
        return block_chunk

    def open(self):
        '''Open the log file'''
        if self.is_open(): return
        self.file = open(self.filename,'rb')
        # Check file open format
        byte_chunk = self.file.read(len(self.formatter.file_open_chunk()))
        if not self.formatter.valid_endian(byte_chunk):
            raise ValueError('Incorrect endian word:',byte_chunk)
        return

    def is_open(self):
        '''Check if log file is open'''
        if self.file is None: return False
        return not self.file.closed
    
    def next_block(self):
        '''Load and return next serial data block'''
        # Read bytes
        block_bytes = self.next_block_bytes()
        # Catch end of data
        if block_bytes is None:
            return None
        # Parse block
        block_desc = self.formatter.parse_block(block_bytes)
        # Return
        return block_desc
    
    def close(self):
        '''Close the log file'''
        self.file.close()
        self.file = None
        return

def print_log(filename):
    '''Print (to terminal) the data blocks from a serial log file'''
    if not os.path.isfile(filename):
        print("Error: File does not exist: '%s'" % filename)
    loader = DataLoader(filename)
    while True:
        block = loader.next_block()
        if block is None: break
        print('%%%%%%%%%%%%%%%%')
        print('  Block type:',block['block_type'])
        if block['block_type']=='file':
            print(block)
        if block['block_type']=='data':
            print('  Data type:  ',block['data_type'])
            print('  Size:       ',len(block['data']))
            if len(block['data'])>10:
                print('  Start bytes:',block['data'][:10])
                print('  End bytes:  ',block['data'][-10:])
            else:
                print('  Bytes:      ',block['data'])
    return



'''
A module to assist loading serial data logs for debugging
'''
from __future__ import absolute_import
from datetime import datetime
from math import sqrt

from .dataloader import DataLoader
from larpix import Controller
from larpix import Packet

class LogAnalyzer(DataLoader):
    '''Analyzer of LArPix serial log transmissions'''
    def __init__(self, filename=None):
        '''Constructor'''
        DataLoader.__init__(self, filename)
        # Merge packets split across transmissions?
        self._stitch_transmissions = False
        self._unused_bytes = bytes()

    def next_transmission(self):
        '''Parse next set of packets in controller transmission log'''
        block_desc = self.next_block()
        if block_desc is None:
            return None
        if block_desc['block_type'] == 'data':
            parse_bytes = bytes()
            if self._stitch_transmissions and len(self._unused_bytes)>0:
                # Use leftover bytes from previous transmission
                parse_bytes += self._unused_bytes
                self._unused_bytes = bytes()
            # Get bytes from this transmission
            parse_bytes += bytes(block_desc['data'])
            if self._stitch_transmissions:
                # Remove extra bytes for use in next transmission
                #   FIXME: assumes all packets are 10 bytes in length!
                n_extra_bytes = (len(parse_bytes) % 10)
                if n_extra_bytes != 0:
                    self._unused_bytes = parse_bytes[-n_extra_bytes:]
                    parse_bytes = parse_bytes[:-n_extra_bytes]
            packets = Controller.parse_input(parse_bytes)
            block_desc['packets'] = packets
        return block_desc

    def dump_log(self):
        '''Dump entire log to terminal'''
        while True:
            next_trans = self.next_transmission()
            if next_trans is None:
                break
            self.print_transmission(next_trans)
        return

    def check_parity(self, transmission):
        '''Check parities of all packets in transmission'''
        npackets_bad_parity = 0
        if transmission['block_type'] != 'data': return 0
        return sum(1 for pack in transmission['packets']
                   if not pack.has_valid_parity())

    def check_fifo(self, transmission):
        '''Returns tuple of number of half full fifos and full fifos in transmission'''
        half_flags = 0
        full_flags = 0
        if transmission['block_type'] != 'data': return (0,0)
        for packet in transmission['packets']:
            if packet.packet_type == Packet.DATA_PACKET:
                half_flags += int(packet.fifo_half_flag)
                full_flags += int(packet.fifo_full_flag)
        return (half_flags, full_flags)

    def adc_report(self, interval_step=100000, max_read=None, return_list=False):
        '''
        Print min, max, avg channel adcs for all packets with good parity in log
        Returns list of adc values for each channel if ``fast=False``
        Channel id is ``chip_id * 32 + channel``
        '''
        print('========= ADC Report ========================')
        npackets_total = 0
        npackets_bad_parity_total = 0
        npackets_good_total = {}
        adc_min_total = {}
        adc_max_total = {}
        adc_avg_total = {}
        adc_ssq_total = {} # sum of squares
        adc_total = {}
        npackets_interval = 0
        npackets_bad_parity_interval = 0
        npackets_good_interval = {}
        adc_min_interval = {}
        adc_max_interval = {}
        adc_avg_interval = {}
        adc_ssq_interval = {}
        while True:
            next_trans = self.next_transmission()
            if next_trans is None:
                break
            if next_trans['block_type'] != 'data': continue
            if next_trans['data_type'] != 'read': continue
            if len(next_trans['packets']) <= 0: continue
            npackets_total += len(next_trans['packets'])
            npackets_interval += len(next_trans['packets'])
            for packet in next_trans['packets']:
                if packet.packet_type != Packet.DATA_PACKET:
                    continue
                adc = packet.dataword
                chipid = packet.chipid
                channel = packet.channel_id
                channelid = chipid * 32 + channel
                # Remove packets with bad parity
                if packet.has_valid_parity() is False:
                    npackets_bad_parity_total += 1
                    npackets_bad_parity_interval += 1
                    continue
                # Store adc info for good packets
                if not channelid in npackets_good_total:
                    npackets_good_total[channelid] = 1
                    adc_max_total[channelid] = adc
                    adc_min_total[channelid] = adc
                    adc_avg_total[channelid] = adc
                    adc_ssq_total[channelid] = adc * adc
                    adc_total[channelid] = [adc]
                else:
                    npackets_good_total[channelid] += 1
                    adc_max_total[channelid] = max(adc, adc_max_total[channelid])
                    adc_min_total[channelid] = min(adc, adc_min_total[channelid])
                    adc_avg_total[channelid] += float(adc - adc_avg_total[channelid])/npackets_good_total[channelid] # running avg
                    adc_ssq_total[channelid] += adc * adc
                    adc_total[channelid].append(adc)
                if not channelid in npackets_good_interval:
                    npackets_good_interval[channelid] = 1
                    adc_max_interval[channelid] = adc
                    adc_min_interval[channelid] = adc
                    adc_avg_interval[channelid] = adc
                    adc_ssq_interval[channelid] = adc * adc
                else:
                    npackets_good_interval[channelid] += 1
                    adc_max_interval[channelid] = max(adc, adc_max_interval[channelid])
                    adc_min_interval[channelid] = min(adc, adc_min_interval[channelid])
                    adc_avg_interval[channelid] += float(adc - adc_avg_interval[channelid])/npackets_good_interval[channelid] # running avg
                    adc_ssq_interval[channelid] += adc * adc
            if not max_read is None and npackets_total > max_read:
                # Stop at max
                break
            if npackets_interval > interval_step:
                # Interval Report
                print('  N_packets, N_badparity: %d %d' % (
                    npackets_interval,
                    npackets_bad_parity_interval))
                print('  Chip, Channel, Good packets, Min ADC, Max ADC, Avg ADC, Std Dev ADC')
                for channelid in npackets_good_interval.keys():
                    print('  %4d, %7d, %12d, %7d, %7d, %7.2f, %11.2f' % (
                            channelid // 32,
                            channelid % 32,
                            npackets_good_interval[channelid],
                            adc_min_interval[channelid],
                            adc_max_interval[channelid],
                            adc_avg_interval[channelid],
                            sqrt(float(adc_ssq_interval[channelid])/npackets_good_interval[channelid] - adc_avg_interval[channelid]**2)))
                    del npackets_good_interval[channelid]
                    del adc_min_interval[channelid]
                    del adc_max_interval[channelid]
                    del adc_avg_interval[channelid]
                    del adc_ssq_interval[channelid]
                npackets_interval = 0
                npackets_bad_parity_interval = 0
        # Total Report
        print('== Summary of ADC values ==')
        print('  Total number of packets: %d' % npackets_total)
        print('  Packets with bad parity: %d' % npackets_bad_parity_total)
        print('  Channel breakdown:')
        print('  Chip, Channel, Good packets, Min ADC, Max ADC, Avg ADC, Std Dev ADC')
        mean_std_dev = 0
        n = 0
        for channelid in sorted(npackets_good_total.keys()):
            print('  %4d, %7d, %12d, %7d, %7d, %7.2f, %11.2f' % (
                    channelid // 32,
                    channelid % 32,
                    npackets_good_total[channelid],
                    adc_min_total[channelid],
                    adc_max_total[channelid],
                    adc_avg_total[channelid],
                    sqrt(float(adc_ssq_total[channelid])/npackets_good_total[channelid] - adc_avg_total[channelid]**2)))
        if return_list:
            return adc_total

    def parity_report(self, interval_step=100000):
        '''Check parities of all packets in log, and report status'''
        print('========= Parity Report =====================')
        npackets_total = 0
        npackets_bad_parity_total = 0
        npackets_interval = 0
        npackets_bad_parity_interval = 0
        while True:
            next_trans = self.next_transmission()
            if next_trans is None:
                break
            if next_trans['block_type'] != 'data': continue
            if next_trans['data_type'] != 'read': continue
            if len(next_trans['packets']) <= 0: continue
            npackets_bad_parity = self.check_parity(next_trans)
            npackets_total += len(next_trans['packets'])
            npackets_bad_parity_total += npackets_bad_parity
            npackets_interval += len(next_trans['packets'])
            npackets_bad_parity_interval += npackets_bad_parity
            if npackets_interval > interval_step:
                # Interval Report
                print('  N_packets, N_badparity, Bad Fraction: %d %d %f' % (
                    npackets_interval,
                    npackets_bad_parity_interval,
                    npackets_bad_parity_interval/float(npackets_interval)))
                npackets_interval = 0
                npackets_bad_parity_interval = 0
        # Total Report
        print('== Summary of Parity Check ==')
        print('  Total number of packets: %d' % npackets_total)
        print('  Packets with bad parity: %d' % npackets_bad_parity_total)
        if npackets_total > 0:
            print('                 Fraction: %f' % (npackets_bad_parity_total/
                                                     float(npackets_total)))
        return

    def fifo_report(self, interval_step=100000):
        '''Check fifo flags of all packets in log, and report status'''
        print('========= FIFO Report =====================')
        npackets_total = 0
        npackets_half_fifo_total = 0
        npackets_full_fifo_total = 0
        npackets_interval = 0
        npackets_half_fifo_interval = 0
        npackets_full_fifo_interval = 0
        while True:
            next_trans = self.next_transmission()
            if next_trans is None:
                break
            if next_trans['block_type'] != 'data': continue
            if next_trans['data_type'] != 'read': continue
            if len(next_trans['packets']) <= 0: continue
            n_half, n_full = self.check_fifo(next_trans)
            npackets_total += len(next_trans['packets'])
            npackets_half_fifo_total += n_half
            npackets_full_fifo_total += n_full
            npackets_interval += len(next_trans['packets'])
            npackets_half_fifo_interval += n_half
            npackets_full_fifo_interval += n_full
            if npackets_interval > interval_step:
                # Interval Report
                print('  N_packets, N_half, N_full: %d %d %d' % (
                    npackets_interval,
                    npackets_half_fifo_interval,
                    npackets_full_fifo_interval))
                npackets_interval = 0
                npackets_half_fifo_interval = 0
                npackets_full_fifo_interval = 0
        # Total Report
        print('== Summary of FIFO Check ==')
        print('  Total number of packets: %d' % npackets_total)
        print('  Packets with half full FIFO: %d' % npackets_half_fifo_total)
        print('  Packets with full FIFO: %d' % npackets_full_fifo_total)
        return

    @classmethod
    def print_transmission(cls, block, show_packets=True):
        '''Print transmission to terminal'''
        tfmt = '%Y-%m-%d %H:%M:%S.%f'
        print('%%%%%%%%%%%%%%%% Next Block %%%%%%%%%%%%%%%%')
        print(' Block Type: %s' % block['block_type'])
        if block['block_type']=='file':
            print('  Data version: %d.%d' % (block['major_version'],
                                             block['minor_version']))
            utc_dt = datetime.utcfromtimestamp(block['starttime'])
            loc_dt = datetime.fromtimestamp(block['starttime'])
            print('  Time, UTC:    %s' % utc_dt.strftime(tfmt))
            print('     (Local):  (%s)' % loc_dt.strftime(tfmt))
        if block['block_type']=='data':
            print('  Data type:    %s' % block['data_type'])
            utc_dt = datetime.utcfromtimestamp(block['time'])
            loc_dt = datetime.fromtimestamp(block['time'])
            print('  Time, UTC:    %s' % utc_dt.strftime(tfmt))
            print('     (Local):  (%s)' % loc_dt.strftime(tfmt))
            print('  Size [bytes]: %d' % len(block['data']))
            print('  N(packets):   %d' % len(block['packets']))
            if show_packets and (len(block['packets']) > 0):
                print('  Packets:')
                for packet in block['packets']:
                    print('   %s' % str(packet))
        return




from collections import OrderedDict
import warnings
import time
import json
import math
import networkx as nx
from copy import copy

from . import configs
from .key import Key
from .chip import Chip
from .configuration import Configuration_v1, Configuration_v2
from .packet import Packet_v1, Packet_v2, PacketCollection
from . import bitarrayhelper as bah

class Controller(object):
    '''
    Controls a collection of LArPix Chip objects.

    Reading data:

    The specific interface for reading data is selected by specifying
    the ``io`` attribute. These objects all have
    similar behavior for reading in new data. On initialization, the
    object will discard any LArPix packets sent from ASICs. To begin
    saving incoming packets, call ``start_listening()``.
    Data will then build up in some form of internal register or queue.
    The queue can be emptied with a call to ``read()``,
    which empties the queue and returns a list of Packet objects that
    were in the queue. The ``io`` object will still be listening for
    new packets during and after this process. If the queue/register
    fills up, data may be discarded/lost. To stop saving incoming
    packets and retrieve any packets still in the queue, call
    ``stop_listening()``. While the Controller is listening,
    packets can be sent using the appropriate methods without
    interrupting the incoming data stream.

    With the the LArPix v2 asic, a network io structure is used. To keep track of
    this network and to insure the proper bring-up procedure is followed. The
    controller has a ``network`` attribute. This is an ordered dict with multiple
    layers corresponding to the chip key layers. It assumes that each hydra network
    is distinct on a given io channel. On a given io channel, the hydra network is
    represented as three directed graphs ``miso_us``, ``miso_ds``, and ``mosi``.::

        controller.network[1] # Represents all networks on io group 1
        controller.network[1][1] # Represents the network on io group 1, io channel 1
        controller.network[1][1]['mosi'] # directed graph representing mosi links

    Nodes within the network are added automatically whenever you add a chip to the
    controller::

        list(controller.network[1][1]['mosi'].nodes()) # chip ids of known mosi nodes
        list(controller.network[1][1]['miso_us'].nodes()) # chip ids of known miso_us nodes
        list(controller.network[1][1]['miso_ds'].nodes()) # chip ids of known miso_ds nodes
        controller.network[1][1]['miso_us'].nodes[5] # attributes associated with node 5
        controller.add_chip('1-1-6', version=2, root=True)
        list(controller.network[1][1]['miso_us'].nodes()) # [5, 6]


    The 'root' node is used to indicate which chips should be brought up first when
    initializing the network.

    Accessing connections between nodes is done via edges::

        list(controller.network[1][1]['miso_us'].edges()) # []
        controller.add_network_link(1,1,'miso_us',(6,5),0) # link chips 6 -> 5 in the miso_us graph via chip 6's uart channel 0
        list(controller.network[1][1]['miso_us'].edges()) # [(6,5)] direction indicates the data flow direction
        controller.network[1][1]['miso_us'].edges[(6,5)] # attributes associated with link ('uart': 0)

    In order to set up a 'working' network, you'll need to add the proper links in
    the miso_ds and mosi graphs as well.::

        controller.add_network_link(1,1,'miso_ds',(5,6),2)
        controller.add_network_link(1,1,'mosi',(6,5),0)
        controller.add_network_link(1,1,'mosi',(5,6),2)

    External systems (e.g. the fpga you are using to communicate with the chips) are
    conventionally called ``'ext0', 'ext1', ...`` and shouldn't be forgotten in the
    network configuration! These nodes are used to determine the miso/mosi
    configuration for the linked chips.::

        controller.add_network_link(1,1,'mosi',('ext',6),1) # link chip 6 to 'ext' via chip 6's uart channel 0
        controller.add_network_link(1,1,'miso_ds',(6,'ext'),1) # same for miso_ds
        list(controller.network[1][1]['miso_ds'].edges())
        list(controller.network[1][1]['mosi'].edges())

    To then initialize a component on the network, run `controller.init_network(1,1,6)`.
    This will modify the correct registers of chip 1-1-6 and send the configuration
    commands to the system.::

        controller.init_network(1,1,6)
        controller.init_network(1,1,5)

    Keep in mind, the order in which you initialize the network is important (e.g.
    chip 6 needs to be initialized before chip 5 since it is upstream from chip 6).

    You may also reset a particular node with `reset_network(<io_group>,<io_channel>,<chip_id>)`
    which resets the network configuration of a particular node.::

        controller.reset_network(1,1,5)
        controller.reset_network(1,1,6)

    This is all rather tedious to do manually, so there a few shortcuts to make declaring and
    configuring the network easier. If you don't want to initialize or reset each
    node in order, you may allow the controller to figure out the proper order
    (based on the root nodes) and configure all of the chips.::

        controller.init_network(1,1) # configure the 1-1 network
        controller.reset_network(1,1) # revert the 1-1 network to it's original state

    In the event that you need to access the chip keys of a network in the order
    of their depth within the network, use::

        controller.get_network_keys(1,1) # get a list of chip keys starting with the root node and descending into the network
        controller.get_network_keys(1,1,root_first_traversal=False) # get a list of chip keys starting with the chips deepest in the network and ascending

    You may also specify a network configuration file and use this to create all
    of the necessary network links and chips. See
    <https://larpix-control.readthedocs.io/en/stable/api/configs/controller.html>
    for how to create one of these files. This is the generally recommended means of
    creating and loading a Hydra io network.::

        # as an example for bringing up a hydra network from a config file
        controller.load('<network config file>.json') # load your network and chips
        for io_group, io_channels in controller.network.items():
            for io_channel in io_channels:
                controller.init_network(io_group, io_channel) # update configurations and write to chips

    Properties and attributes:

    - ``chips``: the ``Chip`` objects that the controller controls
    - ``reads``: list of all the PacketCollections that have been sent
      back to this controller. PacketCollections are created by
      ``run``, ``write_configuration``, ``read_configuration``,
      ``multi_write_configuration``, ``multi_read_configuration``, and
      ``store_packets``.
    - ``network``: a collection of networkx directed graph objects representing
      the miso_us, miso_ds, and mosi connections between chips (not applicable
      for v1 asics)

    '''
    network_names = ('miso_us', 'miso_ds', 'mosi')

    def __init__(self):
        self.chips = OrderedDict()
        self.network = OrderedDict()
        self.reads = []
        self.nreads = 0
        self.io = None
        self.logger = None

    def __getitem__(self, key):
        '''
        Retrieve the Chip object that this Controller associated with the key.
        A key can either be a valid keystring, ``larpix.Key`` object or a
        ``tuple``/``list``. If it is a ``tuple``/``list``, the chip with a key
        of ``Key(*key)`` will be retrieved.

        Raises a ``ValueError`` if no chip is found.

        E.g.::

            controller[1,1,2] == controller['1-1-2'] # access chip with key 1-1-2
            controller[1,1,2] == controller[Key('1-1-2')]

        '''
        return self._get_chip(key)

    def _get_chip(self, key):
        try:
            if isinstance(key, (str,Key)):
                return self.chips[key]
            elif isinstance(key, (tuple,list)):
                return self.chips[Key(*key)]
            raise KeyError
        except KeyError:
            raise ValueError('Could not find chip using key <{}> '.format(key))

    def add_chip(self, chip_key, version=2, config=None, root=False):
        '''
        Add a specified chip to the Controller chips.

        :param chip_key: chip key to specify unique chip

        :param version: asic version of chip

        :param config: configuration to assign chip when creating (otherwise uses default)

        :param root: specifies if this is a root node as used by the hydra io network

        :returns: ``Chip`` that was added

        '''
        if chip_key in self.chips:
            raise KeyError('chip with key {} already exists!'.format(chip_key))
        key = Key(chip_key)
        io_group, io_channel = key.io_group, key.io_channel
        self.chips[key] = Chip(chip_key=chip_key, version=version)
        self.add_network_node(io_group, io_channel, self.network_names, key.chip_id, root)

        if not config is None:
            self[chip_key].config = config
        return self.chips[chip_key]

    def add_network_link(self, io_group, io_channel, network_name, chip_ids, uart):
        '''
        Adds a link within the specified network (``mosi``, ``miso_ds``, or ``miso_us``)
        on the specified uart. Directionality is important: first key represents
        the tail (where packets are coming from) and the second is the head
        (where packets are going). The uart channel refers to the tail's uart
        channel (for ``'miso_*'``) and to the head's uart channel for (``'mosi'``).
        Links can be formed with non-key objects to allow for network
        communications with objects that are not larpix chips (e.g. an external
        fpga).

        For example to specify a miso downstream link between chips ``'1-1-3'``
        and ``'1-1-2'`` on chip ``'1-1-3'``'s uart 0, the arguments should be::

            controller.add_network_link('miso_ds', 1, 1, (3,2), 0)

        This represents a link where chip ``'1-1-3'`` will transfer packets out of
        its uart 0 to a destination of chip ``'1-1-2'`` (unspecified uart).

        No network validation is performed, so use with caution!

        :param network_name: ``str`` from ``('miso_us','miso_ds','mosi')``

        :param io_group: io group to add network link in

        :param io_channel: io channel to add network link in

        :param chip_ids: ``tuple`` of two chip ids to link (order is important -- first is 'tail' second is 'head', data flows from the 'tail' to the 'head' in all networks)

        :param uart: ``int`` referring to which uart channel this link occurs over, for ``'miso_*'`` networks the uart channel refers to the uart on the sender (i.e. the 'tail' of the edge in the directed graph), but for ``'mosi'`` networks the uart channel refers to the uart on the receiver (i.e. the 'head' of the edge in the directed graph)

        '''
        if not chip_ids[0] in self.network[io_group][io_channel][network_name]:
            self.add_network_node(io_group, io_channel, network_name, chip_ids[0])
        if not chip_ids[1] in self.network[io_group][io_channel][network_name]:
            self.add_network_node(io_group, io_channel, network_name, chip_ids[1])
        self.network[io_group][io_channel][network_name].add_edge(*chip_ids, uart=uart)

    def add_network_node(self, io_group, io_channel, network_names, chip_id, root=False):
        '''
        Adds a node to the specified network (``mosi``, ``miso_ds``, or ``miso_us``)
        with no links. Generally, this is not needed as network nodes are added
        automatically with the ``controller.add_chip()`` method. 'Special'
        (i.e. non-key) nodes can be specified to insert nodes that do not
        represent larpix chips (e.g. an external fpga). A node can be declared as
        a root node, which is used as a flag to select the starting node during
        initialization.

        :param network_names: list of networks or name of single network to create a node in

        :param io_group: io group to create node within

        :param io_channel: io channel to create node within

        :param chip_id: chip id to associate with node (note that all chip ids must be unique on a network)

        :param root: specifies if this node should be a root node (default=False)

        '''
        self._create_network(io_group, io_channel, network_names)

        if isinstance(network_names, str):
            self.network[io_group][io_channel][network_names].add_node(chip_id, root=root)
        else:
            for name in network_names:
                self.network[io_group][io_channel][name].add_node(chip_id, root=root)

    def get_network_ids(self, io_group, io_channel, root_first_traversal=True):
        '''
        Returns a list of chip ids in order of depth within the miso_us network

        :param io_group: io group of network

        :param io_channel: io channel of network

        :param root_first_traversal: ``True`` to traverse network starting from root nodes then increasing in network depth, ``False`` to traverse network starting from nodes furthest from root nodes and then decreasing in network depth

        '''
        subnetwork = self.network[io_group][io_channel]['miso_us']
        ordered_ids = []

        chip_ids = [chip_id for chip_id in subnetwork.nodes() if subnetwork.nodes[chip_id]['root']]

        collected_chips = set(chip_ids)

        while chip_ids:
            for chip_id in chip_ids:
                ordered_ids.append(chip_id)
                collected_chips.add(chip_id)
            chip_ids = [link[1] for link in subnetwork.out_edges(chip_ids) if not link[1] in collected_chips]
        if root_first_traversal:
            return ordered_ids
        return ordered_ids[::-1]

    def get_network_keys(self, io_group, io_channel, root_first_traversal=True):
        '''
        Returns a list of chip ids in order of depth within the miso_us network

        :param io_group: io group of network

        :param io_channel: io channel of network

        :param root_first_traversal: ``True`` to traverse network starting from root nodes then increasing in network depth, ``False`` to traverse network starting from nodes furthest from root nodes and then decreasing in network depth

        '''
        chip_ids = self.get_network_ids(io_group, io_channel, root_first_traversal=root_first_traversal)
        return [Key(io_group, io_channel, chip_id) for chip_id in chip_ids if isinstance(chip_id, int)]

    def _create_network(self, io_group, io_channel, network_names):
        '''
        Creates specified network if they don't already exist

        :param network_names: list of networks or name of single network to create

        :param io_group: io group to create node within

        :param io_channel: io channel to create node within

        '''
        if not io_group in self.network.keys():
            self.network[io_group] = OrderedDict()
        if not io_channel in self.network[io_group].keys():
            self.network[io_group][io_channel] = {}
        if isinstance(network_names, str):
            if not network_names in self.network[io_group][io_channel].keys():
                self.network[io_group][io_channel][network_names] = nx.DiGraph()
        else:
            for name in network_names:
                if not name in self.network[io_group][io_channel].keys():
                    self.network[io_group][io_channel][name] = nx.DiGraph()

    def remove_chip(self, chip_key):
        '''
        Remove a specified chip from the Controller chips.

        :param chip_key: chip key to specify unique chip

        '''
        chip_key = Key(chip_key)
        io_channel, io_group = chip_key.io_channel, chip_key.io_group
        del self.chips[chip_key]

        for network_name in self.network_names:
            self.network[io_group][io_channel][network_name].remove_node(chip_key.chip_id)

    def load(self, filename):
        '''
        Loads the specified file that describes the chip ids and IO network

        :param filename: File path to configuration file

        '''
        system_info = configs.load(filename, 'controller')
        if system_info['asic_version'] == 1:
            print('loading v1 controller...')
            return self.load_controller(filename)
        if system_info['asic_version'] == 2:
            print('loading v2 network...')
            return self.load_network(filename)

    def load_network(self, filename):
        '''
        Loads the specified file using hydra io network configuration format

        After loading the network, running
        ``controller.init_network(<io group>, <io channel>)`` will configure the
        chips with specified chip ids in the order declared in file

        :param filename: File path to configuration file

        '''
        system_info = configs.load(filename)
        inherited_data = ('miso_us_uart_map', 'miso_ds_uart_map', 'mosi_uart_map')
        orig_chips = copy(self.chips)
        orig_network = copy(self.network)
        for chip in copy(self.chips):
            self.remove_chip(chip) # clear chips and network
        try:
            def propogate_inherited_values(network_spec, value_keys):
                dict_to_return = dict()
                for value_key in value_keys:
                    dict_to_return[value_key] = network_spec[value_key]

                for key, group_spec in network_spec.items():
                    if key in value_keys:
                        continue
                    dict_to_return[key] = group_spec

                    for value_key in value_keys:
                        if not value_key in group_spec:
                            dict_to_return[key][value_key] = dict_to_return[value_key]

                    for subkey, channel_spec in network_spec[key].items():
                        if subkey in value_keys:
                            continue
                        for value_key in value_keys:
                            if not value_key in channel_spec:
                               dict_to_return[key][subkey][value_key] = dict_to_return[key][value_key]

                        for idx, node_spec in enumerate(channel_spec['nodes']):
                            for value_key in value_keys:
                                if not value_key in node_spec:
                                   dict_to_return[key][subkey]['nodes'][idx][value_key] = dict_to_return[key][subkey][value_key]

                return dict_to_return

            self.chips = OrderedDict()
            full_network_spec = propogate_inherited_values(system_info['network'], inherited_data)

            # create nodes + chip objects first
            for key, group_spec in full_network_spec.items():
                if key in inherited_data:
                    continue
                io_group = key
                for subkey, channel_spec in full_network_spec[io_group].items():
                    if subkey in inherited_data:
                        continue
                    io_channel = subkey
                    for node_spec in channel_spec['nodes']:
                        io_group = int(io_group)
                        io_channel = int(io_channel)
                        chip_id = node_spec['chip_id']
                        if isinstance(chip_id, int):
                            # create new chip object
                            chip_key = Key(io_group, io_channel, chip_id)
                            root = False
                            if 'root' in node_spec and node_spec['root']:
                                root = True
                            self.add_chip(chip_key, version=2, root=root)
                        else:
                            # dummy node
                            root = False
                            if 'root' in node_spec and node_spec['root']:
                                root = True
                            self.add_network_node(io_group, io_channel, self.network_names, chip_id, root)

            # then create miso_us network
            for key, group_spec in full_network_spec.items():
                if key in inherited_data:
                    continue
                io_group = key
                for subkey, channel_spec in full_network_spec[io_group].items():
                    if subkey in inherited_data:
                        continue
                    io_channel = subkey
                    for node_spec in channel_spec['nodes']:
                        io_group = int(io_group)
                        io_channel = int(io_channel)
                        chip_id = node_spec['chip_id']
                        subnetwork = self.network[io_group][io_channel]

                        if 'miso_us' in node_spec:
                            try:
                                for idx, uart in enumerate(node_spec['miso_us_uart_map']):
                                    link = (chip_id, node_spec['miso_us'][idx])
                                    if link[1] is None:
                                        continue
                                    self.add_network_link(io_group, io_channel, 'miso_us', link, uart)
                            except KeyError:
                                raise KeyError('miso_us_uart_map unspecified for {}-{}-{}'.format(io_group,io_channel,chip_id))

            # then create miso_ds network
            for key, group_spec in full_network_spec.items():
                if key in inherited_data:
                    continue
                io_group = key
                for subkey, channel_spec in full_network_spec[io_group].items():
                    if subkey in inherited_data:
                        continue
                    io_channel = subkey
                    for node_spec in channel_spec['nodes']:
                        io_group = int(io_group)
                        io_channel = int(io_channel)
                        chip_id = node_spec['chip_id']
                        subnetwork = self.network[io_group][io_channel]

                        if 'miso_ds' in node_spec:
                            try:
                                for idx, uart in enumerate(node_spec['miso_ds_uart_map']):
                                    link = (chip_id, node_spec['miso_ds'][idx])
                                    if link[1] is None:
                                        continue
                                    self.add_network_link(io_group, io_channel, 'miso_ds', link, uart)
                            except KeyError:
                                raise KeyError('miso_us_uart_map unspecified for {}-{}-{}'.format(io_group,io_channel,chip_id))
                        elif subnetwork['miso_us'].in_edges(chip_id):
                            try:
                                for link in subnetwork['miso_us'].in_edges(chip_id):
                                    other_chip_id = link[0]
                                    other_spec = [spec for spec in channel_spec['nodes'] if spec['chip_id'] == other_chip_id][0]
                                    uart = node_spec['miso_ds_uart_map'][other_spec['miso_us'].index(chip_id)]
                                    link = (chip_id, other_chip_id)
                                    self.add_network_link(io_group, io_channel, 'miso_ds', link, uart)
                            except KeyError:
                                raise KeyError('miso_ds_uart_map unspecified for {}-{}-{}'.format(io_group,io_channel,chip_id))

            # finally, create mosi network
            for key, group_spec in full_network_spec.items():
                if key in inherited_data:
                    continue
                io_group = key
                for subkey, channel_spec in full_network_spec[io_group].items():
                    if subkey in inherited_data:
                        continue
                    io_channel = subkey
                    for node_spec in channel_spec['nodes']:
                        io_group = int(io_group)
                        io_channel = int(io_channel)
                        chip_id = node_spec['chip_id']
                        subnetwork = self.network[io_group][io_channel]

                        if 'mosi' in node_spec:
                            try:
                                for idx, uart in enumerate(node_spec['mosi_uart_map']):
                                    link = (chip_id, node_spec['mosi'][idx])
                                    if link[1] is None:
                                        continue
                                    self.add_network_link(io_group, io_channel, 'mosi', link, uart)
                            except KeyError:
                                raise KeyError('mosi_uart_map unspecified for {}-{}-{}'.format(io_group,io_channel,chip_id))
                        elif subnetwork['miso_us'].in_edges(chip_id) or subnetwork['miso_ds'].in_edges(chip_id):
                            try:
                                # create links for existing miso_us connections
                                if subnetwork['miso_us'].in_edges(chip_id):
                                    for link in subnetwork['miso_us'].in_edges(chip_id):
                                        other_chip_id = link[0]
                                        other_spec = [spec for spec in channel_spec['nodes'] if spec['chip_id'] == other_chip_id][0]
                                        uart = node_spec['mosi_uart_map'][other_spec['miso_us'].index(chip_id)]
                                        link = (other_chip_id, chip_id)
                                        self.add_network_link(io_group, io_channel, 'mosi', link, uart)

                                # create links for existing miso_ds connections
                                if subnetwork['miso_ds'].in_edges(chip_id):
                                    for link in subnetwork['miso_ds'].in_edges(chip_id):
                                        other_chip_id = link[0]
                                        other_spec = [spec for spec in channel_spec['nodes'] if spec['chip_id'] == other_chip_id][0]
                                        uart = None
                                        if 'miso_ds' in other_spec:
                                            uart = node_spec['mosi_uart_map'][other_spec['miso_ds'].index(chip_id)]
                                        else:
                                            # look up via ds uart channel map
                                            ds_uart = subnetwork['miso_ds'].edges[link]['uart']
                                            uart = node_spec['mosi_uart_map'][node_spec['miso_us_uart_map'].index(ds_uart)]
                                        link = (other_chip_id, chip_id)
                                        self.add_network_link(io_group, io_channel, 'mosi', link, uart)
                            except KeyError:
                                raise KeyError('mosi_uart_map unspecified for {}-{}-{}'.format(io_group,io_channel,chip_id))

        except Exception as err:
            self.chips = orig_chips
            self.network = orig_network
            raise err

        return system_info['name']

    def load_controller(self, filename):
        '''
        Loads the specified file using the basic key, chip format

        The chip key is the Controller access key that gets communicated to/from
        the io object when sending and receiving packets.

        :param filename: File path to configuration file

        '''
        system_info = configs.load(filename)
        chips = OrderedDict()
        for chip_keystring in system_info['chip_list']:
            chip_key = Key(str(chip_keystring))
            chips[chip_key] = Chip(chip_key=chip_key, version=1)
        self.chips = chips
        return system_info['name']

    def init_network_and_verify(self, io_group=1, io_channel=1, chip_id=None, timeout=0.2, retries=10):
        '''
        Runs init network, verifying that registers are updated properly at each step
        Will exit if any chip network configuration cannot be set correctly after
        ``retries`` attempts

        '''
        if chip_id is None:
            chip_ids = self.get_network_ids(io_group, io_channel, root_first_traversal=True)
            for chip_id in chip_ids:
                ok,diff = self.init_network_and_verify(io_group, io_channel, chip_id=chip_id, timeout=timeout, retries=retries)
                if not ok:
                    return ok,diff
            return ok,dict()

        keys_to_verify = list()
        keys_to_verify.append(Key(io_group,io_channel,chip_id))
        for us_link in self.network[io_group][io_channel]['miso_us'].in_edges(chip_id):
            if not isinstance(us_link[0], int):
                continue
            parent_chip_id = us_link[0]
            keys_to_verify.append(Key(io_group, io_channel, parent_chip_id))

        print('init',io_group,io_channel,chip_id)
        self.init_network(io_group, io_channel, chip_id)
        ok,diff = self.verify_network(keys_to_verify, timeout=timeout)
        for _ in range(retries):
            if ok: return ok,diff
            for chip_key in diff:
                self.write_configuration(chip_key, diff[chip_key].keys())
            ok,diff = self.verify_registers([
                (chip_key, list(diff[chip_key].keys())) for chip_key in diff
            ])
        if not ok:
            print('Failed to verify network nodes {}'.format(keys_to_verify))
        return ok,diff

    def init_network(self, io_group=1, io_channel=1, chip_id=None):
        '''
        Configure a Hydra io node specified by chip_id, if none are specified,
        load complete network
        To configure a Hydra io network, the following steps are followed:

         - Enable miso_us of parent chip
         - Write chip_id to chip on io group + io channel (assuming chip id 1)
         - Write enable_miso_downstream to chip chip_id
         - Write enable_mosi to chip chip_id
         - Write enable_miso_differential to chip chip_id

        '''
        subnetwork = self.network[io_group][io_channel]
        if chip_id is None:
            chip_ids = self.get_network_ids(io_group, io_channel, root_first_traversal=True)
            for chip_id in chip_ids:
                self.init_network(io_group, io_channel, chip_id=chip_id)
            return

        packets = []
        chip_key = None
        if isinstance(chip_id, int):
            chip_key = Key(io_group, io_channel, chip_id)

        # Enable miso_upstream on parent chips
        for us_link in subnetwork['miso_us'].in_edges(chip_id):
            if not isinstance(us_link[0], int):
                continue
            parent_chip_id = us_link[0]
            parent_chip_key = Key(io_group, io_channel, parent_chip_id)
            parent_uart = subnetwork['miso_us'].edges[us_link]['uart']
            self[parent_chip_key].config.enable_miso_upstream[parent_uart] = 1
            packets += self[parent_chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_upstream'])

        if chip_key:
            # Only modify chip configuration if node points to a chip object

            # Write chip_id to specified chip
            self[chip_key].config.chip_id = chip_key.chip_id
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['chip_id'])
            packets[-1].chip_id = 1

            # Enable miso_downstream, mosi on chip
            self[chip_key].config.enable_miso_downstream = [0]*4
            for ds_link in subnetwork['miso_ds'].out_edges(chip_id):
                ds_uart = subnetwork['miso_ds'].edges[ds_link]['uart']
                self[chip_key].config.enable_miso_downstream[ds_uart] = 1
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_downstream'])

            self[chip_key].config.enable_mosi = [0]*4
            for mosi_link in subnetwork['mosi'].in_edges(chip_id):
                mosi_uart = subnetwork['mosi'].edges[mosi_link]['uart']
                self[chip_key].config.enable_mosi[mosi_uart] = 1
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_mosi'])

        self.send(packets)

    def reset_network(self, io_group=1, io_channel=1, chip_id=None):
        '''
        Resets the hydra IO network
        To reset a Hydra io network node, the following steps are followed:

         - Write enable_mosi [1,1,1,1] to chip
         - Write enable_miso_downstream [0,0,0,0] to chip
         - Write chip_id 1 to chip
         - Disable miso_us channel of parent chip

        If no chip_id is specified, network is reset in reverse order (starting
        from chips with no outgoing miso_us connections, working up the miso_us
        tree)

        '''
        subnetwork = self.network[io_group][io_channel]
        if chip_id is None:
            chip_keys = self.get_network_keys(io_group, io_channel, root_first_traversal=False)
            for chip_key in chip_keys:
                self.reset_network(io_group, io_channel, chip_id=chip_key.chip_id)
            return

        packets = []
        chip_key = None
        if isinstance(chip_id, int):
            # Only modify chip configuration if node points to a chip object
            chip_key = Key(io_group, io_channel, chip_id)

            # Enable mosi
            self[chip_key].config.enable_mosi = [1]*4
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_mosi'])

            # Disable miso_downstream
            self[chip_key].config.enable_miso_downstream = [0]*4
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_downstream'])

            # Write default chip_id to specified chip
            self[chip_key].config.chip_id = 1
            packets += self[chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['chip_id'])

        # Disable miso_upstream on parent chips
        for us_link in subnetwork['miso_us'].in_edges(chip_id):
            if not isinstance(us_link[0], int):
                continue
            parent_chip_id = us_link[0]
            parent_chip_key = Key(io_group, io_channel, parent_chip_id)
            parent_uart = subnetwork['miso_us'].edges[us_link]['uart']
            self[parent_chip_key].config.enable_miso_upstream[parent_uart] = 0
            packets += self[parent_chip_key].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_upstream'])

        self.send(packets)


    def send(self, packets):
        '''
        Send the specified packets to the LArPix ASICs.

        '''
        timestamp = time.time()
        if self.io:
            self.io.send(packets)
        else:
            warnings.warn('no IO object exists, no packets sent', RuntimeWarning)
        if self.logger:
            self.logger.record(packets, direction=self.logger.WRITE)

    def start_listening(self):
        '''
        Listen for packets to arrive.

        '''
        if self.io:
            self.io.start_listening()
        else:
            warnings.warn('no IO object exists, you have done nothing', RuntimeWarning)

    def stop_listening(self):
        '''
        Stop listening for new packets to arrive.

        '''
        if self.io:
            return self.io.stop_listening()
        else:
            warnings.warn('no IO object exists, you have done nothing', RuntimeWarning)

    def read(self):
        '''
        Read any packets that have arrived and return (packets,
        bytestream) where bytestream is the bytes that were received.

        The returned list will contain packets that arrived since the
        last call to ``read`` or ``start_listening``, whichever was most
        recent.

        '''
        timestamp = time.time()
        packets = []
        bytestream = b''
        if self.io:
            packets, bytestream = self.io.empty_queue()
        else:
            warnings.warn('no IO object exists, no packets will be received', RuntimeWarning)
        if self.logger:
            self.logger.record(packets, direction=self.logger.READ)
        return packets, bytestream

    def write_configuration(self, chip_key, registers=None, write_read=0,
                            message=None, connection_delay=0.2):
        '''
        Send the configurations stored in chip.config to the LArPix
        ASIC.

        By default, sends all registers. If registers is an int, then
        only that register is sent. If registers is a string, then
        the sent registers will set by looking at the configuration
        register map. If registers is an iterable, then
        all of the registers in the iterable are sent.

        If write_read == 0 (default), the configurations will be sent
        and the current listening state will not be affected. If the
        controller is currently listening, then the listening state
        will not change and the value of write_read will be ignored. If
        write_read > 0 and the controller is not currently listening,
        then the controller will listen for ``write_read`` seconds
        beginning immediately before the packets are sent out, read the
        io queue, and save the packets into the ``reads`` data member.
        Note that the controller will only read the queue once, so if a
        lot of data is expected, you should handle the reads manually
        and set write_read to 0 (default).

        '''
        chip = self[chip_key]
        if registers is None:
            registers = list(range(chip.config.num_registers))
        elif isinstance(registers, int):
            registers = [registers]
        elif isinstance(registers, str):
            registers = list(chip.config.register_map[registers])
        else:
            pass
        if message is None:
            message = 'configuration write'
        else:
            message = 'configuration write: ' + message
        packets = chip.get_configuration_write_packets(registers)
        already_listening = False
        if self.io:
            already_listening = self.io.is_listening
        mess_with_listening = write_read != 0 and not already_listening
        if mess_with_listening:
            self.start_listening()
            time.sleep(connection_delay)
            stop_time = time.time() + write_read
        self.send(packets)
        if mess_with_listening:
            sleep_time = stop_time - time.time()
            if sleep_time > 0:
                time.sleep(sleep_time)
            packets, bytestream = self.read()
            self.stop_listening()
            self.store_packets(packets, bytestream, message)

    def read_configuration(self, chip_key, registers=None, timeout=1,
                           message=None, connection_delay=0.2):
        '''
        Send "configuration read" requests to the LArPix ASIC.

        By default, request all registers. If registers is an int, then
        only that register is reqeusted. If registers is an iterable,
        then all of the registers in the iterable are requested.

        If the controller is currently listening, then the requests
        will be sent and no change to the listening state will occur.
        (The value of ``timeout`` will be ignored.) If the controller
        is not currently listening, then the controller will listen
        for ``timeout`` seconds beginning immediately before the first
        packet is sent out, and will save any received packets in the
        ``reads`` data member.

        '''
        chip = self[chip_key]
        if registers is None:
            registers = list(range(chip.config.num_registers))
        elif isinstance(registers, int):
            registers = [registers]
        elif isinstance(registers, str):
            registers = chip.config.register_map[registers]
        else:
            pass
        if message is None:
            message = 'configuration read'
        else:
            message = 'configuration read: ' + message
        packets = chip.get_configuration_read_packets(registers)
        already_listening = False
        if self.io:
            already_listening = self.io.is_listening
        if not already_listening:
            self.start_listening()
            time.sleep(connection_delay)
            stop_time = time.time() + timeout
        self.send(packets)
        if not already_listening:
            sleep_time = stop_time - time.time()
            if sleep_time > 0:
                time.sleep(sleep_time)
        packets, bytestream = self.read()
        self.stop_listening()
        self.store_packets(packets, bytestream, message)

    def multi_write_configuration(self, chip_reg_pairs, write_read=0,
                                  message=None, connection_delay=0.2):
        '''
        Send multiple write configuration commands at once.

        ``chip_reg_pairs`` should be a list/iterable whose elements are
        an valid arguments to ``Controller.write_configuration``,
        excluding the ``write_read`` argument. Just like in the single
        ``Controller.write_configuration``, setting ``write_read > 0`` will
        have the controller read data during and after it writes, for
        however many seconds are specified.

        Examples:

        These first 2 are equivalent and write the full configurations

        >>> controller.multi_write_configuration([chip_key1, chip_key2, ...])
        >>> controller.multi_write_configuration([(chip_key1, None), chip_key2, ...])

        These 2 write the specified registers for the specified chips
        in the specified order

        >>> controller.multi_write_configuration([(chip_key1, 1), (chip_key2, 2), ...])
        >>> controller.multi_write_configuration([(chip_key1, range(10)), chip_key2, ...])

        '''
        if message is None:
            message = 'multi configuration write'
        else:
            message = 'multi configuration write: ' + message
        packets = []
        for chip_reg_pair in chip_reg_pairs:
            if not isinstance(chip_reg_pair, tuple):
                chip_reg_pair = (chip_reg_pair, None)
            chip_key, registers = chip_reg_pair
            chip = self[chip_key]
            if registers is None:
                registers = list(range(chip.config.num_registers))
            elif isinstance(registers, int):
                registers = [registers]
            else:
                pass
            one_chip_packets = chip.get_configuration_write_packets(registers)
            packets.extend(one_chip_packets)
        already_listening = False
        if self.io:
            already_listening = self.io.is_listening
        mess_with_listening = write_read != 0 and not already_listening
        if mess_with_listening:
            self.start_listening()
            time.sleep(connection_delay)
            stop_time = time.time() + write_read
        self.send(packets)
        if mess_with_listening:
            sleep_time = stop_time - time.time()
            if sleep_time > 0:
                time.sleep(sleep_time)
            #time.sleep(stop_time - time.time())
            packets, bytestream = self.read()
            self.stop_listening()
            self.store_packets(packets, bytestream, message)

    def multi_read_configuration(self, chip_reg_pairs, timeout=1,
                                 message=None, connection_delay=0.2):
        '''
        Send multiple read configuration commands at once.

        ``chip_reg_pairs`` should be a list/iterable whose elements are
        chip keys (to read entire configuration) or (chip_key, registers)
        tuples to read only the specified register(s). Registers could
        be ``None`` (i.e. all), an ``int`` for that register only, or an
        iterable of ints.

        Examples:

        These first 2 are equivalent and read the full configurations

        >>> controller.multi_read_configuration([chip_key1, chip_key2, ...])
        >>> controller.multi_read_configuration([(chip_key1, None), chip_key2, ...])

        These 2 read the specified registers for the specified chips
        in the specified order

        >>> controller.multi_read_configuration([(chip_key1, 1), (chip_key2, 2), ...])
        >>> controller.multi_read_configuration([(chip_key1, range(10)), chip_key2, ...])

        '''
        if message is None:
            message = 'multi configuration read'
        else:
            message = 'multi configuration read: ' + message
        packets = []
        for chip_reg_pair in chip_reg_pairs:
            if not isinstance(chip_reg_pair, tuple):
                chip_reg_pair = (chip_reg_pair, None)
            chip_key, registers = chip_reg_pair
            chip = self[chip_key]
            if registers is None:
                registers = list(range(chip.config.num_registers))
            elif isinstance(registers, int):
                registers = [registers]
            else:
                pass
            one_chip_packets = chip.get_configuration_read_packets(registers)
            packets += one_chip_packets
        already_listening = False
        if self.io:
            already_listening = self.io.is_listening
        if not already_listening:
            self.start_listening()
            time.sleep(connection_delay)
            stop_time = time.time() + timeout
        self.send(packets)
        if not already_listening:
            sleep_time = stop_time - time.time()
            if sleep_time > 0:
                time.sleep(sleep_time)
        packets, bytestream = self.read()
        self.stop_listening()
        self.store_packets(packets, bytestream, message)

    def run(self, timelimit, message):
        '''
        Read data from the LArPix ASICs for the given ``timelimit`` and
        associate the received Packets with the given ``message``.

        '''
        sleeptime = min(0.1, timelimit)
        self.start_listening()
        start_time = time.time()
        packets = []
        bytestreams = []
        while time.time() - start_time < timelimit:
            time.sleep(sleeptime)
            read_packets, read_bytestream = self.read()
            packets.extend(read_packets)
            bytestreams.append(read_bytestream)
        self.stop_listening()
        data = b''.join(bytestreams)
        self.store_packets(packets, data, message)

    def verify_registers(self, chip_key_register_pairs, timeout=1, connection_delay=0.02):
        '''
        Read chip configuration from specified chip and registers and return ``True`` if the
        read chip configuration matches the current configuration stored in chip instance.
        ``chip_key`` is a single chip key.

        :param chip_key_register_pair: a ``list`` of key register pairs as documented in ``controller.multi_read_configuration``

        :param timeout: set how long to wait for response in seconds (optional)

        :returns: 2-``tuple`` of a ``bool`` representing if all registers match and a ``dict`` representing all differences. Differences are specified as ``{<chip_key>: {<register>: (<expected>, <read>)}}``

        '''
        return_value = True
        different_fields = {}
        registers = {}
        for chip_key, chip_registers in chip_key_register_pairs:
            if not isinstance(chip_registers, (list,tuple,range)):
                if chip_key in registers:
                    registers[chip_key] += [chip_registers]
                else:
                    registers[chip_key] = [chip_registers]
            else:
                if chip_key in registers:
                    registers[chip_key] += list(chip_registers)
                else:
                    registers[chip_key] = list(chip_registers)
        self.multi_read_configuration(chip_key_register_pairs, timeout=timeout, connection_delay=connection_delay)
        configuration_data = dict([
            (chip_key, dict([
                (register,(None, None))
                for register in chip_registers]))
            for chip_key, chip_registers in registers.items()])
        for packet in self.reads[-1]:
            packet_key = packet.chip_key
            if (hasattr(packet,'CONFIG_READ_PACKET') and packet.packet_type == packet.CONFIG_READ_PACKET):
                register_address = packet.register_address
                if packet_key in configuration_data and register_address in configuration_data[packet_key]:
                    configuration_data[packet_key][register_address] = (None, packet.register_data)
                    
        for chip_key in registers.keys():
            expected_data = dict()
            if self[chip_key].asic_version == 1:
                expected_data = dict([(register_address, bah.touint(bits)) for register_address, bits in enumerate(self[chip_key].config.all_data())])
            else:
                expected_data = dict([(register_address, bah.touint(bits, endian=Packet_v2.endian)) for register_address, bits in enumerate(self[chip_key].config.all_data())])

            for register in set(registers[chip_key]):
                configuration_data[chip_key][register] = (expected_data[register], configuration_data[chip_key][register][1])
                if not configuration_data[chip_key][register][0] == configuration_data[chip_key][register][1]:
                    return_value = False
                else:
                    del configuration_data[chip_key][register]
            if not len(configuration_data[chip_key]):
                del configuration_data[chip_key]
        return (return_value, configuration_data)

    def verify_configuration(self, chip_keys=None, timeout=1):
        '''
        Read chip configuration from specified chip(s) and return ``True`` if the
        read chip configuration matches the current configuration stored in chip instance.
        ``chip_keys`` can be a single chip key, a list of chip keys, or ``None``. If
        ``chip_keys`` is ``None`` all chips will be verified.

        Also returns a dict containing the values of registers that are different
        (read register, stored register)

        :param chip_keys: ``list`` of chip_keys to verify

        :param timeout: how long to wait for response in seconds

        :returns: 2-``tuple`` with same format as ``controller.verify_registers``

        '''
        if chip_keys is None:
            chip_keys = self.chips.keys()
        if isinstance(chip_keys,(str,Key)):
            chip_keys = [chip_keys]
        chip_key_register_pairs = [(chip_key, range(self[chip_key].config.num_registers)) for chip_key in chip_keys]
        return self.verify_registers(chip_key_register_pairs, timeout=timeout)

    def verify_network(self, chip_keys=None, timeout=1):
        '''
        Read chip network configuration from specified chip(s) and return ``True``
        if the read chip configurations matches
        Only valid for v2 chips.

        :param chip_keys: ``list`` of chip_keys to verify or singe chip_key to verify

        :param timeout: how long to wait for response in seconds

        :returns: 2-``tuple`` with same format as ``controller.verify_registers``

        '''
        if not chip_keys:
            chip_keys = self.chips.keys()
        if isinstance(chip_keys,(str,Key)):
            chip_keys = [chip_keys]
        network_registers = list(Configuration_v2.register_map['chip_id']) + \
            list(Configuration_v2.register_map['enable_mosi']) + \
            list(Configuration_v2.register_map['enable_miso_upstream']) + \
            list(Configuration_v2.register_map['enable_miso_downstream']) + \
            list(Configuration_v2.register_map['enable_miso_differential'])
        chip_key_register_pairs = [(chip_key, network_registers) for chip_key in chip_keys]
        return self.verify_registers(chip_key_register_pairs)

    def enable_analog_monitor(self, chip_key, channel):
        '''
        Enable the analog monitor on a single channel on the specified chip.
        Note: If monitoring a different chip, call disable_analog_monitor first to ensure
        that the monitor to that chip is disconnected.
        '''
        chip = self[chip_key]
        if chip.asic_version == 1:
            chip.config.disable_analog_monitor()
            chip.config.enable_analog_monitor(channel)
            self.write_configuration(chip_key, chip.config.csa_monitor_select_addresses)
        elif chip.asic_version == 2:
            chip.config.csa_monitor_select = [0]*chip.config.num_channels
            chip.config.csa_monitor_select[channel] = 1
            self.write_configuration(chip_key, chip.config.register_map['csa_monitor_select'])
        else:
            raise RuntimeError('chip has invalid asic version')
        return

    def disable_analog_monitor(self, chip_key=None, channel=None):
        '''
        Disable the analog monitor for a specified chip and channel, if none are specified
        disable the analog monitor for all chips in self.chips and all channels
        '''
        if chip_key is None:
            for chip in self.chips:
                self.disable_analog_monitor(chip_key=chip, channel=channel)
        else:
            chip = self[chip_key]
            if chip.asic_version == 1:
                chip.config.disable_analog_monitor()
                self.write_configuration(chip_key, chip.config.csa_monitor_select_addresses)
            elif chip.asic_version == 2:
                if not channel is None:
                    chip.config.csa_monitor_select[channel] = 0
                else:
                    chip.config.csa_monitor_select = [0]*64
                self.write_configuration(chip_key, chip.config.register_map['csa_monitor_select'])
            else:
                raise RuntimeError('chip has invalid asic version')
        return

    def enable_testpulse(self, chip_key, channel_list, start_dac=255):
        '''
        Prepare chip for pulsing - enable testpulser and set a starting dac value for
        specified chip/channel
        '''
        chip = self[chip_key]
        if chip.asic_version == 1:
            chip.config.disable_testpulse()
            chip.config.enable_testpulse(channel_list)
            chip.config.csa_testpulse_dac_amplitude = start_dac
            self.write_configuration(chip_key, chip.config.csa_testpulse_enable_addresses +
                                     [chip.config.csa_testpulse_dac_amplitude_address])
        elif chip.asic_version == 2:
            chip.config.csa_testpulse_enable = [1]*chip.config.num_channels
            for channel in channel_list:
                chip.config.csa_testpulse_enable[channel] = 0
            chip.config.csa_testpulse_dac = start_dac
            self.write_configuration(chip_key, list(chip.config.register_map['csa_testpulse_dac']) + list(chip.config.register_map['csa_testpulse_enable']))
        else:
            raise RuntimeError('chip has invalid asic version')
        return

    def issue_testpulse(self, chip_key, pulse_dac, min_dac=0, read_time=0.1):
        '''
        Reduce the testpulser dac by ``pulse_dac`` and write_read to chip for
        ``read_time`` seconds

        '''
        chip = self[chip_key]
        if chip.asic_version == 1:
            chip.config.csa_testpulse_dac_amplitude -= pulse_dac
            if chip.config.csa_testpulse_dac_amplitude < min_dac:
                raise ValueError('Minimum DAC exceeded')
            self.write_configuration(chip_key, [chip.config.csa_testpulse_dac_amplitude_address],
                                     write_read=read_time)
        elif chip.asic_version == 2:
            if chip.config.csa_testpulse_dac - pulse_dac < min_dac:
                raise ValueError('Minimum DAC exceeded')
            try:
                chip.config.csa_testpulse_dac -= pulse_dac
            except ValueError:
                raise ValueError('Minimum DAC exceeded')
            self.write_configuration(chip_key, chip.config.register_map['csa_testpulse_dac'], write_read=read_time)
        return self.reads[-1]

    def disable_testpulse(self, chip_key=None, channel_list=None):
        '''
        Disable testpulser for specified chip/channels. If none specified, disable for
        all chips/channels
        '''
        if chip_key is None:
            for chip_key in self.chips.keys():
                self.disable_testpulse(chip_key=chip_key, channel_list=channel_list)
        if channel_list is None:
            channel_list = range(self[chip_key].config.num_channels)
            self.disable_testpulse(chip_key=chip_key, channel_list=channel_list)
        else:
            chip = self[chip_key]
            if chip.asic_version == 1:
                chip.config.disable_testpulse(channel_list)
                self.write_configuration(chip_key, chip.config.csa_testpulse_enable_addresses)
            elif chip.asic_version == 2:
                for channel in channel_list:
                    chip.config.csa_testpulse_enable[channel] = 1
                self.write_configuration(chip_key, chip.config.register_map['csa_testpulse_enable'])
            else:
                raise RuntimeError('chip has invalid asic version')
        return

    def disable(self, chip_key=None, channel_list=None):
        '''
        Update channel mask to disable specified chips/channels. If none specified,
        disable all chips/channels
        '''
        if chip_key is None:
            for chip_key in self.chips.keys():
                self.disable(chip_key=chip_key, channel_list=channel_list)
        else:
            chip = self[chip_key]
            if channel_list is None:
                channel_list = range(chip.config.num_channels)
            if chip.asic_version == 1:
                chip.config.disable_channels(channel_list)
                self.write_configuration(chip_key, Configuration_v1.channel_mask_addresses)
            elif chip.asic_version == 2:
                for channel in channel_list:
                    chip.config.channel_mask[channel] = 1
                    self.write_configuration(chip_key, chip.config.register_map['channel_mask'])
            else:
                raise RuntimeError('chip has invalid asic version')

    def enable(self, chip_key=None, channel_list=None):
        '''
        Update channel mask to enable specified chips/channels. If none specified,
        enable all chips/channels
        '''
        if chip_key is None:
            for chip_key in self.chips.keys():
                self.enable(chip_key=chip_key, channel_list=channel_list)
        else:
            chip = self[chip_key]
            if channel_list is None:
                channel_list = range(chip.config.num_channels)
            if chip.asic_version == 1:
                chip.config.enable_channels(channel_list)
                self.write_configuration(chip_key, Configuration_v1.channel_mask_addresses)
            elif chip.asic_version == 2:
                for channel in channel_list:
                    chip.config.channel_mask[channel] = 0
                self.write_configuration(chip_key, chip.config.register_map['channel_mask'])
            else:
                raise RuntimeError('chip has invalid asic version')

    def store_packets(self, packets, data, message):
        '''
        Store the packets in ``self`` and in ``self.chips``

        '''
        new_packets = PacketCollection(packets, data, message)
        new_packets.read_id = self.nreads
        self.nreads += 1
        self.reads.append(new_packets)
        #self.sort_packets(new_packets)

    def sort_packets(self, collection):
        '''
        Sort the packets in ``collection`` into each chip in ``self.chips``

        '''
        by_chip_key = collection.by_chip_key()
        for chip_key in by_chip_key.keys():
            if chip_key in self.chips.keys():
                chip = self[chip_key]
                chip.reads.append(by_chip_key[chip_key])
            elif not self._test_mode:
                print('Warning chip key {} not in chips.'.format(chip_key))

    def save_output(self, filename, message):
        '''Save the data read by each chip to the specified file.'''
        data = {}
        data['reads'] = [collection.to_dict() for collection in self.reads]
        data['chips'] = [repr(chip) for chip in self.chips.values()]
        data['message'] = message
        with open(filename, 'w') as outfile:
            json.dump(data, outfile, indent=4,
                    separators=(',',':'), sort_keys=True)



'''
This module contains io configuration files used by io classes to look up
detector components based on the ``io_group`` and ``io_channel`` contained within
chip keys. See the larpix core documentation for more details on ``larpix.Key``
objects.

JSON formatting
---------------

The format is a standard JSON file structured as follows:

.. parsed-literal::
    {
        "_config_type": "io",
        "io_class": "<name of io class config should be used with>",
        "io_group": [
            [<io group number>, <assoc. value to be used by io class>],
            ...
        ]
    }

Field description
-----------------

The ``"_config_type": "io"`` field is used in loading for validation (so that
you don't accidentally try to load a ``chip`` config into your io class. This
will always be "io" for io confiugration files.

The ``"io_class": "<name>"`` field is used to specify the io class that the
configuration is compatible with. Examples are provided for each built-in io
class.

The ``"io_group": [[<group #>, <io class spec.>], ...]`` list is a list of pairs
used to create a map between the ``io_group`` number and the internal
representation used by the io class. E.g. the MultiZMQ_IO uses the IP address to
identify the ``io_group``.

'''
pass



from bitarray import bitarray
import copy

from larpix import Packet_v2
from larpix import bitarrayhelper as bah

packet_types = {
    'data':Packet_v2.DATA_PACKET,
    'test':Packet_v2.TEST_PACKET,
    'write':Packet_v2.CONFIG_WRITE_PACKET,
    'read':Packet_v2.CONFIG_READ_PACKET
    }
shared_fields = ['chip_id','downstream_marker','parity']
data_fields = ['channel_id','timestamp','dataword','trigger_type','local_fifo','shared_fifo']
write_read_fields = ['register_address','register_data']

def test_basic():
    p1 = Packet_v2()
    assert p1.bits == bitarray('0'*64)
    assert p1 == Packet_v2()
    assert p1.bytes() == b'\x00'*8

    p2 = Packet_v2(b'\x01' + b'\x00'*7)
    assert p2 != p1
    assert p2.bits == bitarray('10000000') + bitarray('0'*56)

def test_field_assignment():
    '''
    Tests that you can properly get, set, export, and from_dict each simple data
    attribute

    '''

    p = Packet_v2()

    def test_field(packet, field_name):
        print('testing {}'.format(field_name))
        p = copy.deepcopy(packet)

        assert getattr(p,field_name) == 0
        assert bah.touint(p.bits[getattr(p,field_name+'_bits')], endian=p.endian) == 0

        setattr(p,field_name,1)
        assert getattr(p,field_name) == 1
        assert bah.touint(p.bits[getattr(p,field_name+'_bits')], endian=p.endian) == 1
        assert str(p) # just to make sure there are no errors in printing the string

        p_dict = p.export()
        print(p_dict)
        assert field_name in p_dict
        assert p_dict[field_name] == 1

        p_dict[field_name] = 0
        p2 = Packet_v2(p.bytes())
        setattr(p2, field_name, 0)
        p_dict['bits'] = p2.bits.to01()
        p.from_dict(p_dict)
        assert getattr(p,field_name) == 0
        assert bah.touint(p.bits[getattr(p,field_name+'_bits')], endian=p.endian) == 0


    for type_name, packet_type in packet_types.items():
        print('testing packet type {}'.format(type_name))
        p.packet_type = packet_type
        for field in shared_fields:
            test_field(p, field)
        if type_name == 'data':
            for field in data_fields:
                test_field(p, field)
        elif type_name == 'write':
            for field in write_read_fields:
                test_field(p, field)
        elif type_name == 'read':
            for field in write_read_fields:
                test_field(p, field)

def test_chip_key():
    p = Packet_v2()

    def assert_chip_key(packet, key, group, channel, chip_id):
        assert packet.chip_key == key
        assert packet.io_group == group
        assert packet.io_channel == channel
        assert packet.chip_id == chip_id
        assert bah.touint(p.bits[p.chip_id_bits], endian=p.endian) == chip_id

    # check default values
    assert_chip_key(p,None,None,None,0)

    # check chip key assignment
    p.chip_key = '1-2-3'
    assert_chip_key(p,'1-2-3',1,2,3)

    # check io group assignment
    p.io_group = 4
    assert_chip_key(p,'4-2-3',4,2,3)

    # check io channel assignment
    p.io_channel = 5
    assert_chip_key(p,'4-5-3',4,5,3)

    # check chip id assignment
    p.chip_id = 6
    assert_chip_key(p,'4-5-6',4,5,6)

def test_fifo_diagnostics():
    p = Packet_v2()

    # no fifo diagnositics => local/shared fifo events do nothing
    assert p.fifo_diagnostics_enabled == False
    p.local_fifo_events = 1
    p.shared_fifo_events = 2
    p_dict = p.export()
    assert 'local_fifo_events' not in p_dict.keys()
    assert 'shared_fifo_events' not in p_dict.keys()
    assert p.local_fifo_events == None
    assert p.shared_fifo_events == None
    assert p.timestamp == 0

    # fifo diagnostics => local/shared fifo modify corresponding bits
    p.fifo_diagnostics_enabled = True
    p.local_fifo_events = 1
    p.shared_fifo_events = 2
    p_dict = p.export()
    assert p_dict['local_fifo_events'] == 1
    assert p_dict['shared_fifo_events'] == 2
    assert p.local_fifo_events == 1
    assert p.shared_fifo_events == 2
    assert p.timestamp == 0

    # no fifo diagnostics => local/shared fifo events do nothing again
    p.fifo_diagnostics_enabled = False
    assert p.timestamp == 268566528
    p.local_fifo_events = 7
    p.shared_fifo_events = 8
    assert p.timestamp == 268566528
    assert p.local_fifo_events == None
    assert p.shared_fifo_events == None

def test_parity():
    p = Packet_v2()
    assert p.parity == 0
    assert p.compute_parity() == 1
    assert not p.has_valid_parity()

    p.assign_parity()
    assert p.has_valid_parity()
    assert p.parity == 1

    p.bits[0] = 1
    assert p.parity == 1
    assert p.compute_parity() == 0
    assert not p.has_valid_parity()

    p.assign_parity()
    assert p.has_valid_parity()
    assert p.parity == 0



'''
A python interface to the pacman ZMQ message format.

The pacman ZMQ messages are a raw bytestring with two components: a short
header and N words. It's a bit cumbersome to work with bytestrings in data (and
keep track of endianness / etc), so this module allows you to translate between
native python objects and the pacman message bytestring.

To import this module::

    import larpix.format.pacman_msg_format as pacman_msg_fmt

Access to the message content is provided by the ``parse_msg(msg)`` method which
takes a single pacman message bytestring as an argument, e.g.::

    msg = b'!-----\x01\x00P---------------' # a simple ping response
    data = pacman_msg_fmt.parse_msg(msg) # (('REP',123456,1), [('PONG')])
    msg = b'D-----\x01\x00D\x01------datadata' # a simple data message
    data = pacman_msg_fmt.parse_msg(msg) # (('DATA',123456,1), [('DATA',1,1234,b'datadata')])

The creation of messages can be performed with the inverse method,
``format_msg(msg_type, msg_words)``. Here, the ``msg_type`` is one of ``'DATA'``,
``'REQ'``, or ``'REP'``, and ``msg_words`` is a ``list`` of tuples indicating the word
type (index 0) and the word data (index 1, 2, ...). Word types are specified
by strings that can be found in the ``word_type_table``. The necessary fields
(and primitive types) for the word data is described in the ``word_fmt_table``.
E.g.::

    data = ('REP', [('PONG')]) # simple ping response
    msg = pacman_msg_fmt.format_msg(*data) # b'!----\x00\x01\x00P\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    data = ('DATA', [('DATA',1,1234,b'datadata')])
    msg = pacman_msg_fmt.format_msg(*data) # b'D----\x00\x01\x00D\x01\x00\x00\x00\x00\x00\x00datadata'

To facilitate translating to/from ``larpix-control`` packet objects, you can use
the ``format(pkts, msg_type)`` and ``parse(msg, io_group=None)`` methods. E.g.::

    packet = Packet_v2()
    packet.io_channel = 1
    msg = pacman_msg_fmt.format([packet]) # b'?----\x00\x01\x00D\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

    packets = pacman_msg_fmt.parse(msg, io_group=1) # [Packet_v2(b'\x00\x00\x00\x00\x00\x00\x00\x00')]
    packets[0].io_group # 1

Note that no ``io_group`` data is contained within a pacman message. This means
that when formatting messages, packets' ``io_group`` field is ignored, and when
parsing messages, an ``io_group`` value needs to be specified at the time of
parsing.

'''
import struct
from bidict import bidict
import time

from larpix import Packet_v2, TriggerPacket, SyncPacket, TimestampPacket

HEADER_LEN=8
WORD_LEN=16
MSG_TYPE_DATA=b'D'
MSG_TYPE_REQ=b'?'
MSG_TYPE_REP=b'!'
WORD_TYPE_DATA=b'D'
WORD_TYPE_TRIG=b'T'
WORD_TYPE_SYNC=b'S'
WORD_TYPE_PING=b'P'
WORD_TYPE_WRITE=b'W'
WORD_TYPE_READ=b'R'
WORD_TYPE_TX=WORD_TYPE_DATA
WORD_TYPE_PONG=WORD_TYPE_PING
WORD_TYPE_ERR=b'E'

msg_type_table = bidict([
    ('REQ',MSG_TYPE_REQ),
    ('REP',MSG_TYPE_REP),
    ('DATA',MSG_TYPE_DATA)
    ])
word_type_table = dict(
    REQ=bidict([
        ('PING',WORD_TYPE_PING),
        ('WRITE',WORD_TYPE_WRITE),
        ('READ',WORD_TYPE_READ),
        ('TX',WORD_TYPE_TX)
        ]),
    REP=bidict([
        ('WRITE',WORD_TYPE_WRITE),
        ('READ',WORD_TYPE_READ),
        ('TX',WORD_TYPE_TX),
        ('PONG',WORD_TYPE_PONG),
        ('ERR',WORD_TYPE_ERR)
        ]),
    DATA=bidict([
        ('DATA',WORD_TYPE_DATA),
        ('TRIG',WORD_TYPE_TRIG),
        ('SYNC',WORD_TYPE_SYNC)
        ])
    )

msg_header_fmt = '<cLxH'
msg_header_struct = struct.Struct(msg_header_fmt)

word_fmt_table = dict(
    DATA='<cBLxx8s',
    TRIG='<2cxxL8x',
    SYNC='<2cBxL8x',
    PING='<c15x',
    WRITE='<c3xL4xL',
    READ='<c3xL4xL',
    TX='<cB6x8s',
    PONG='<c15x',
    ERR='<cB14s'
    )
word_struct_table = dict([
    (word_type, struct.Struct(word_fmt))
    for word_type, word_fmt in word_fmt_table.items()
    ])

def format_header(msg_type, msg_words):
    '''
    Generates a header-formatted bytestring of
    message type ``msg_type`` and with ``msg_words``
    specified words

    '''
    return msg_header_struct.pack(
        msg_type_table[msg_type],
        int(time.time()),
        msg_words
        )

def format_word(msg_type, word_type, *data):
    '''
    Generates a word-formatted bytestring of
    word type specified by the ``msg_type`` and
    ``word_type``. The data contained within the word
    should be passed in as additional positional arguments
    in the order that they appear in the word (least significant byte first). E.g. for a data word::

        data_word = format_word('DATA','DATA',<io_channel>,<receipt_timestamp>,<data_word_content>)

    '''
    return word_struct_table[word_type].pack(
        word_type_table[msg_type][word_type],
        *data
        )

def parse_header(msg):
    '''
    Returns a tuple of the data contained in the header::

        (<msg type>, <msg time>, <msg words>)

    '''
    msg_header_data = msg_header_struct.unpack(msg[:HEADER_LEN])
    return (msg_type_table.inv[msg_header_data[0]],) + tuple(msg_header_data[1:])

def parse_word(msg_type, word):
    '''
    Returns a tuple of data contained in word (little endian),
    first item is always the word type

    '''
    word_type = word_type_table[msg_type].inv[word[0:1]]
    return (word_type,) + tuple(word_struct_table[word_type].unpack(word)[1:])

def format_msg(msg_type, msg_words):
    '''
    ``msg_words`` should be a list of tuples that can be unpacked and passed into ``format_word``

    '''
    bytestream = format_header(msg_type, len(msg_words))
    for msg_word in msg_words:
        bytestream += format_word(msg_type, *msg_word)
    return bytestream

def parse_msg(msg):
    '''
    returns a tuple of::

        (<header data>, [<word 0 data>, ...])

    '''
    header = parse_header(msg)
    words = list()
    for idx in range(HEADER_LEN,len(msg),WORD_LEN):
        words.append(parse_word(
            header[0],
            msg[idx:idx+WORD_LEN]
            ))
    return header, words

def _replace_none(obj, attr, default=0):
    return getattr(obj, attr) if getattr(obj, attr) is not None else default

def _packet_data_req(pkt, *args):
    if isinstance(pkt, Packet_v2):
        return ('TX',
                _replace_none(pkt,'io_channel'),
                pkt.bytes())
    return tuple()

def _packet_data_data(pkt, ts_pacman, *args):
    if isinstance(pkt, Packet_v2):
        return ('DATA',
                _replace_none(pkt,'io_channel'),
                pkt.receipt_timestamp if hasattr(pkt,'receipt_timestamp') else ts_pacman,
                pkt.bytes())
    elif isinstance(pkt, SyncPacket):
        return ('SYNC',
                _replace_none(pkt,'sync_type'),
                _replace_none(pkt,'clk_source'),
                _replace_none(pkt,'timestamp'))
    elif isinstance(pkt, TriggerPacket):
        return ('TRIG',
                _replace_none(pkt,'trigger_type'),
                _replace_none(pkt,'timestamp'))

def format(packets, msg_type='REQ', ts_pacman=0):
    '''
    Converts larpix packets into a single PACMAN message.
    The message header is automatically generated.

    Note:: For request messages, this method only formats ``Packet_v2`` objects. For data messages, this method only formats ``Packet_v2``, ``SyncPacket``, and ``TriggerPacket`` objects.

    '''
    get_data = _packet_data_req
    if msg_type == 'DATA':
        get_data = _packet_data_data

    word_datas = list()
    for packet in packets:
        word_data = get_data(packet, ts_pacman)
        if len(word_data) == 0: continue
        word_datas.append(word_data)
    return format_msg(msg_type, word_datas)

def parse(msg, io_group=None):
    '''
    Converts a PACMAN message into larpix packets

    The header is parsed into a ``TimestampPacket``,
    data words are parsed into ``Packet_v2`` objects,
    trigger words are parsed into ``TriggerPacket`` objects,
    and sync words are parsed into ``SyncPacket`` objects.

    '''
    packets = list()
    header, word_datas = parse_msg(msg)
    packets.append(TimestampPacket(timestamp=header[1]))
    packets[0].io_group = io_group
    for word_data in word_datas:
        packet = None
        if word_data[0] in ('TX', 'DATA'):
            packet = Packet_v2(word_data[-1])
            packet.receipt_timestamp = word_data[2]
            packet.io_group = io_group
            packet.io_channel = word_data[1]
        elif word_data[0] is 'TRIG':
            packet = TriggerPacket(trigger_type=word_data[1], timestamp=word_data[2])
            packet.io_group = io_group
        elif word_data[0] is 'SYNC':
            packet = SyncPacket(sync_type=word_data[1], clk_source=word_data[2] & 0x01, timestamp=word_data[3])
            packet.io_group = io_group
        if packet is not None:
            packets.append(packet)
    return packets









import pytest
import json

from larpix import Controller, Configuration_v2, Key
from larpix.io import FakeIO

@pytest.fixture
def network_config(tmpdir):
    '''
    Defines a network of:
        MISO_US:
            3       13
            ^       ^
            2   >   12

        MISO_DS:
            3       13
            v       v
            2   <   12
            v

    '''
    filename = str(tmpdir.join('test_network_conf.json'))
    config_dict = {
        "_config_type": "controller",
        "name": "test",
        "asic_version": 2,
        "network": {
            "1": {
                "1": {
                    "nodes": [
                        {
                            "chip_id": "ext",
                            "root": True,
                            "miso_us": [2, None, None, None]
                        },
                        {
                            "chip_id": 2,
                            "miso_us": [3, None, None, 12],
                        },
                        {
                            "chip_id": 3
                        },
                        {
                            "chip_id": 12,
                            "miso_us": [13, None, None, None]
                        },
                        {
                            "chip_id": 13
                        }
                    ],
                }
            },
            "miso_us_uart_map": [0,1,2,3],
            "miso_ds_uart_map": [2,3,0,1],
            "mosi_uart_map": [2,3,0,1]
        }
    }
    with open(filename,'w') as of:
        json.dump(config_dict, of)
    return filename

@pytest.fixture
def inheriting_network_config(tmpdir, network_config):
    filename = str(tmpdir.join('other_test_network_conf.json'))
    config_dict = {
        "_config_type": "controller",
        "_include": [network_config],
        "name": "other_test",
        "network": {
            "1": {
                "2": {
                    "nodes": [
                        {
                            "chip_id": "ext",
                            "root": True,
                            "miso_us": [None, None, 123, None]
                        },
                        {
                            "chip_id": 123,
                        }
                    ],
                    "miso_us_uart_map": [0,1,2,3],
                    "miso_ds_uart_map": [2,3,0,1],
                }
            },
        }
    }
    with open(filename,'w') as of:
        json.dump(config_dict, of)
    return filename

@pytest.fixture
def network_controller(network_config):
    c = Controller()
    c.load(network_config)
    c.io = FakeIO()
    return c

def test_controller_network(network_controller):
    c = network_controller
    networks = ('miso_us','miso_ds','mosi')
    for chip_key in c.chips:
        for network in networks:
            assert chip_key.chip_id in c.network[chip_key.io_group][chip_key.io_channel][network].nodes()
    us_links = [('ext',2),(2,3),(2,12),(12,13)]
    ds_links = [(3,2),(12,2),(13,12),(2,'ext')]
    mosi_links = [link[::-1] for link in us_links] + [link[::-1] for link in ds_links]
    assert set(c.network[chip_key.io_group][chip_key.io_channel]['miso_us'].edges()) == set(us_links)
    assert set(c.network[chip_key.io_group][chip_key.io_channel]['miso_ds'].edges()) == set(ds_links)
    assert set(c.network[chip_key.io_group][chip_key.io_channel]['mosi'].edges()) == set(mosi_links)

    c.remove_chip('1-1-3')
    assert len(c.chips) == 3
    assert len(c.network[1][1]['miso_us'].edges()) == 3
    assert len(c.network[1][1]['miso_ds'].edges()) == 3
    assert len(c.network[1][1]['mosi'].edges()) == 6
    assert not any(3 in c.network[1][1][network] for network in networks)

    c.add_chip('1-1-3')
    assert len(c.chips) == 4
    assert len(c.network[1][1]['miso_us'].nodes()) == 5
    assert len(c.network[1][1]['miso_ds'].nodes()) == 5
    assert len(c.network[1][1]['mosi'].nodes()) == 5
    assert all(3 in c.network[1][1][network] for network in networks)

    c.add_network_link(1,1,'miso_us',(2,3),0)
    assert len(c.network[1][1]['miso_us'].edges()) == 4
    assert c.network[1][1]['miso_us'].edges[(2,3)]['uart'] == 0
    c.add_network_link(1,1,'miso_ds',(3,2),2)
    assert len(c.network[1][1]['miso_ds'].edges()) == 4
    assert c.network[1][1]['miso_ds'].edges[(3,2)]['uart'] == 2
    c.add_network_link(1,1,'mosi',(3,2),0)
    c.add_network_link(1,1,'mosi',(2,3),2)
    assert len(c.network[1][1]['mosi'].edges()) == 8
    assert c.network[1][1]['mosi'].edges[(3,2)]['uart'] == 0
    assert c.network[1][1]['mosi'].edges[(2,3)]['uart'] == 2

def test_controller_inherit_config(inheriting_network_config):
    c = Controller()
    c.load(inheriting_network_config)

    assert len(c.chips) == 5
    assert len(c.network[1][1]['mosi'].nodes) == 5
    assert len(c.network[1][2]['mosi'].nodes) == 2

def test_controller_init(network_controller):
    c = network_controller
    assert len(c.chips) == 4
    assert len(c.network[1][1]) == 3
    assert len(c.network[1][1]['miso_us'].edges()) == 4
    assert len(c.network[1][1]['miso_ds'].edges()) == 4
    assert len(c.network[1][1]['mosi'].edges()) == 8
    assert ('ext', 2) in c.network[1][1]['mosi'].in_edges(2)

    c.init_network(1,1,2)
    assert c['1-1-2'].config.chip_id == 2
    assert c['1-1-2'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-2'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-2'].config.enable_mosi == [1,0,1,1]

    c.init_network(1,1,3)
    assert c['1-1-2'].config.enable_miso_upstream == [1,0,0,0]
    assert c['1-1-3'].config.chip_id == 3
    assert c['1-1-3'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-3'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-3'].config.enable_mosi == [0,0,1,0]
    assert c.io.sent[-1][-1] == c['1-1-3'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_mosi'])[0]
    assert c.io.sent[-1][-2] == c['1-1-3'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_downstream'])[0]
    chip_id_config_packet = c['1-1-3'].get_configuration_write_packets(registers=Configuration_v2.register_map['chip_id'])[0]
    chip_id_config_packet.chip_id = 1
    assert c.io.sent[-1][-3] == chip_id_config_packet
    assert c.io.sent[-1][-4] == c['1-1-2'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_upstream'])[0]

    c.init_network(1,1)
    assert c['1-1-2'].config.chip_id == 2
    assert c['1-1-2'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-2'].config.enable_mosi == [1,0,1,1]
    assert c['1-1-2'].config.enable_miso_upstream == [1,0,0,1]
    assert c['1-1-3'].config.chip_id == 3
    assert c['1-1-3'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-3'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-3'].config.enable_mosi == [0,0,1,0]
    assert c['1-1-12'].config.chip_id == 12
    assert c['1-1-12'].config.enable_miso_upstream == [1,0,0,0]
    assert c['1-1-12'].config.enable_miso_downstream == [0,1,0,0]
    assert c['1-1-12'].config.enable_mosi == [1,1,0,0]
    assert c['1-1-13'].config.chip_id == 13
    assert c['1-1-13'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-13'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-13'].config.enable_mosi == [0,0,1,0]

def test_controller_init_complete(network_controller):
    c = network_controller
    c.init_network(1,1)
    assert c['1-1-2'].config.chip_id == 2
    assert c['1-1-2'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-2'].config.enable_mosi == [1,0,1,1]
    assert c['1-1-2'].config.enable_miso_upstream == [1,0,0,1]
    assert c['1-1-3'].config.chip_id == 3
    assert c['1-1-3'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-3'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-3'].config.enable_mosi == [0,0,1,0]
    assert c['1-1-12'].config.chip_id == 12
    assert c['1-1-12'].config.enable_miso_upstream == [1,0,0,0]
    assert c['1-1-12'].config.enable_miso_downstream == [0,1,0,0]
    assert c['1-1-12'].config.enable_mosi == [1,1,0,0]
    assert c['1-1-13'].config.chip_id == 13
    assert c['1-1-13'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-13'].config.enable_miso_downstream == [0,0,1,0]
    assert c['1-1-13'].config.enable_mosi == [0,0,1,0]

def test_controller_reset(network_controller):
    c = network_controller
    c.init_network(1,1)

    c.reset_network(1,1,13)
    assert c['1-1-13'].config.chip_id == 1
    assert c['1-1-13'].config.enable_miso_upstream == [0,0,0,0]
    assert c['1-1-13'].config.enable_miso_downstream == [0,0,0,0]
    assert c['1-1-13'].config.enable_mosi == [1,1,1,1]
    assert c.io.sent[-1][-4] == c['1-1-13'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_mosi'])[0]
    assert c.io.sent[-1][-3] == c['1-1-13'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_downstream'])[0]
    chip_id_config_packet = c['1-1-13'].get_configuration_write_packets(registers=Configuration_v2.register_map['chip_id'])[0]
    chip_id_config_packet.chip_id = 13
    assert c.io.sent[-1][-2] == chip_id_config_packet
    assert c.io.sent[-1][-1] == c['1-1-12'].get_configuration_write_packets(registers=Configuration_v2.register_map['enable_miso_upstream'])[0]

    c.reset_network(1,1)
    for chip_key in c.chips:
        print(chip_key)
        assert c[chip_key].config.enable_miso_upstream == [0,0,0,0]
        assert c[chip_key].config.chip_id == 1
        assert c[chip_key].config.enable_miso_downstream == [0,0,0,0]
        assert c[chip_key].config.enable_mosi == [1,1,1,1]

def test_controller_network_traversal(network_controller):
    c = network_controller

    keys = c.get_network_keys(1,1,root_first_traversal=True)
    assert len(keys) == 4
    assert keys[0] == '1-1-2'
    assert set(keys[1:3]) == set(['1-1-3','1-1-12'])
    assert keys[3] == '1-1-13'

    keys = c.get_network_keys(1,1,root_first_traversal=False)
    assert len(keys) == 4
    assert keys[0] == '1-1-13'
    assert set(keys[1:3]) == set(['1-1-3','1-1-12'])
    assert keys[3] == '1-1-2'




'''
Tests for basic logging functionality

'''
from __future__ import print_function
import pytest
from larpix.larpix import Packet, Controller, Chip
from larpix.io.fakeio import FakeIO
from larpix.logger.stdout_logger import StdoutLogger

def test_enable():
    logger = StdoutLogger(buffer_length=1)
    logger.enable()
    assert logger.is_enabled()
    logger.record(['test'])
    assert len(logger._buffer) == 1

def test_disable():
    logger = StdoutLogger()
    logger.disable()
    assert not logger.is_enabled()
    logger.record(['test'])
    assert len(logger._buffer) == 0

def test_flush():
    logger = StdoutLogger(buffer_length=5)
    logger.enable()
    logger.record(['test'])
    assert len(logger._buffer) == 1
    logger.flush()
    assert len(logger._buffer) == 0
    logger.record(['test']*5)
    assert len(logger._buffer) == 5
    logger.record(['test'])
    assert len(logger._buffer) == 0

def test_record():
    logger = StdoutLogger(buffer_length=1)
    logger.enable()
    logger.record(['test'])
    assert logger._buffer[0] == 'Record: test'

@pytest.mark.filterwarnings("ignore:no IO object")
def test_controller_write_capture(capfd, chip):
    controller = Controller()
    controller.logger = StdoutLogger(buffer_length=100)
    controller.logger.enable()
    controller.chips[chip.chip_key] = chip
    controller.write_configuration(chip.chip_key, 0)
    packet = chip.get_configuration_packets(Packet.CONFIG_WRITE_PACKET)[0]
    assert len(controller.logger._buffer) == 1

def test_controller_read_capture(capfd):
    controller = Controller()
    controller.io = FakeIO()
    controller.io.queue.append(([Packet()], b'\x00\x00'))
    controller.logger = StdoutLogger(buffer_length=100)
    controller.logger.enable()
    controller.run(0.1,'test')
    assert len(controller.logger._buffer) == 1



'''
A module for serial data formatting and parsing
'''
from __future__ import absolute_import
import struct

class DataFormatter(object):
    '''Abstract base class for serial data reader/writer'''
    data_version_major = 0
    data_version_minor = 0
    chunk_size = 8 # simplify life by writing in 8 byte words
    endian_test_word = bytearray(b'\x01\x02\x03\x04\x05\x06\x07\x08')
    header_type_by_desc = {'file': int('0001',2),
                           'data': int('0010',2)}
    header_size_by_desc = {'file': 2} # in chunks
    header_type_mask = int('00001111',2)

    @classmethod
    def file_open_chunk(cls):
        '''First chunk in file, to confirm file format is correct'''
        return cls.endian_test_word

    @classmethod
    def number_of_chunks(cls,n_bytes):
        '''Determine number of chunks needed for the given number of bytes'''
        n_chunks = int(n_bytes / cls.chunk_size)
        if (n_bytes % cls.chunk_size) != 0:
            n_chunks += 1  # must pad to catch remaining bytes
        return n_chunks

    @classmethod
    def header_type(cls,header_chunk):
        '''Return the header type, given the first header chunk'''
        header_bits = cls.header_type_mask & header_chunk[0]
        for (key, val) in cls.header_type_by_desc.items():
            if val == header_bits:
                return key
        raise ValueError('Undefined header type:', header_bits)

    @classmethod
    def header_size(cls,header_chunk):
        '''Return the total header size in chunks, given the first chunk'''
        head_type = cls.header_type(header_chunk)
        return cls.header_size_by_desc[head_type]

    @classmethod
    def block_size(cls,header_chunk):
        '''Return the total block size in chunks, given the header chunks'''
        # For base class, only the file header
        return cls.header_size(header_chunk)

    @classmethod
    def data_continued(cls,block_chunk):
        '''Return true if block data is continued in next block'''
        return False

    @classmethod
    def data_version(cls,header_chunk):
        '''Return the data version, given the first file header chunk'''
        return (header_chunk[2],header_chunk[3])


    @classmethod
    def format_file_header(cls, file_head_desc):
        '''Generate a file header'''
        bytes = bytearray([cls.header_type_by_desc['file'], 0,
                           cls.data_version_major, cls.data_version_minor,
                           0,0,0,0])
        return bytes

    @classmethod
    def parse_file_header(cls, file_head_bytes):
        '''Parse a file header'''
        if not (cls.header_type(file_head_bytes) is 'file'):
            raise ValueError('Not a file header')
        file_head_desc = {}
        file_head_desc['block_type'] = 'file'
        (major_version, minor_version) = cls.data_version(file_head_bytes)
        file_head_desc['major_version'] = major_version
        file_head_desc['minor_version'] = minor_version
        return file_head_desc

    @classmethod
    def format_data_block(cls, data_block_desc):
        '''Generate a data block'''
        raise NotImplementedError

    @classmethod
    def parse_data_block(cls, data_block_bytes):
        '''Parse a data block'''
        raise NotImplementedError

    @classmethod
    def get_formatter_by_version(cls, major_version, minor_version):
        '''Find the correct formatter, given a data version'''
        for formatter in DataFormatter.get_all_formatters():
            if (formatter.data_version_major==major_version
                and formatter.data_version_minor==minor_version):
                return formatter()
        raise ValueError('No parser found for data version %d-%d' %
                         (data_major_version,data_minor_version))

    @classmethod
    def get_all_formatters(cls):
        '''Find all formatters, based on class inheritance'''
        formatters = [cls,] + cls.__subclasses__()
        for scls in cls.__subclasses__():
            formatters += scls.__subclasses__()
        return formatters

    @classmethod
    def valid_endian(cls, bytes):
        '''Check bytes to determine if it is the correct endian word'''
        return (bytes == cls.endian_test_word)

    @classmethod
    def format_block(cls, block_desc):
        '''Check block type and delegate to correct formatter'''
        if block_desc['block_type'] is 'file':
            return cls.format_file_header(block_desc)
        raise ValueError('Undefined block type:', block_desc['block_type'])

    @classmethod
    def parse_block(cls, block):
        '''Check block type and delegate to correct parser'''
        if cls.header_type(block) is 'file':
            return cls.parse_file_header(block)
        raise ValueError('Undefined header type:', header_bits)



class DataFormatter_v1_0(DataFormatter):
    '''Explicit data formatter for data version 1.0'''
    data_version_major = 1
    data_version_minor = 0
    data_type_by_desc = {'read': int('01',2),
                         'write': int('10',2),
                         'message': int('11',2)}
    data_desc_by_type = dict([(val,key) for (key,val) in
                              data_type_by_desc.items()])
    data_type_mask = int('00110000',2)
    data_final_flag = 0
    data_continued_flag = 1
    data_continued_mask = int('01000000',2)
    header_size_by_desc = {'file':2,
                           'data':2} # in chunks
    data_block_max_bytes = int('1'*16,2)  # 65 kbytes

    @classmethod
    def data_continued(cls,block_chunk):
        '''Return true if block data is continued in next block'''
        if cls.header_type(block_chunk) is 'data':
            is_cont = (block_chunk[0] & cls.data_continued_mask)>>6
            return (is_cont == cls.data_continued_flag)
        return False

    @classmethod
    def block_size(cls,header_chunk):
        '''Return the total block size in chunks, given the header chunks'''
        if cls.header_type(header_chunk) == 'file':
            # File header has no data payload
            return cls.header_size(header_chunk)
        elif cls.header_type(header_chunk) == 'data':
            return cls.data_block_size(header_chunk)
        raise ValueError('Cannot determine size of unknown block type')

    @classmethod
    def data_block_size(cls,data_header_chunk):
        '''Return the total block size in chunks, given the header chunks'''
        data_size = struct.unpack('<I',data_header_chunk[4:8])[0]
        return (cls.header_size_by_desc['data']
                + cls.number_of_chunks(data_size))

    @classmethod
    def format_file_header(cls, file_head_desc):
        '''Generate a file header'''
        bytes = bytearray([cls.header_type_by_desc['file'], 0,
                           cls.data_version_major, cls.data_version_minor,
                           0,0,0,0])
        bytes += struct.pack("<Q", int(file_head_desc['starttime']*1e6))
        return bytes

    @classmethod
    def parse_file_header(cls, file_head_bytes):
        '''Parse a file header'''
        file_head_desc = DataFormatter.parse_file_header(file_head_bytes)
        stime = (struct.unpack("<Q",file_head_bytes[8:16])[0]) * 1.0e-6
        file_head_desc['starttime'] = stime
        return file_head_desc

    @classmethod
    def format_data_block(cls, data_block_desc):
        '''Generate a data block'''
        bytes = bytearray()
        header_type = cls.header_type_by_desc['data']
        data_type = cls.data_type_by_desc[data_block_desc['data_type']]
        data = data_block_desc['data']
        n_data_bytes = len(data)
        while n_data_bytes >= 0:
            current_data = data
            cur_data_bytes = n_data_bytes
            if cur_data_bytes > cls.data_block_max_bytes:
                # Chop extra large blocks into smaller blocks
                current_data = data[:cls.data_block_max_bytes]
                data = data[cls.data_block_max_bytes:]
                cur_data_bytes = cls.data_block_max_bytes
                data_head_byte_0 = (header_type
                                    | (data_type << 4)
                                    | (cls.data_continued_flag << 6))
                n_data_bytes = len(data)
            else:
                # Process final block
                data = []
                n_data_bytes = -1 # Set to negative to exit loop
                # Mark as last data block
                data_head_byte_0 = (header_type
                                    | (data_type << 4)
                                    | (cls.data_final_flag << 6))
            bytes += (bytearray([data_head_byte_0, 0, 0, 0])
                      + struct.pack("<I", cur_data_bytes))
            bytes += struct.pack("<Q", int(data_block_desc['time']*1e6))
            if cur_data_bytes > 0:
                bytes += bytearray(current_data)
        data_chunks = cls.number_of_chunks(len(bytes))
        n_pad_bytes = (data_chunks * cls.chunk_size) - len(bytes)
        if n_pad_bytes > 0:
            # Pad block to fill chunk_size
            bytes += bytearray([0]*n_pad_bytes)
        return bytes

    @classmethod
    def parse_data_block(cls, data_block):
        '''Parse a data block'''
        data_block_desc = {}
        if not (cls.header_type(data_block) is 'data'):
            raise ValueError('Not a data block')
        data_block_desc['block_type'] = 'data'
        data_type_flag = (data_block[0] & cls.data_type_mask) >> 4
        data_block_desc['data_type'] = cls.data_desc_by_type[data_type_flag]
        head_size = cls.header_size_by_desc['data']
        head_bytes = head_size * cls.chunk_size
        data_len = struct.unpack('<I',data_block[4:8])[0]
        dattime = (struct.unpack("<Q",data_block[8:16])[0]) * 1.0e-6
        data_block_desc['time'] = dattime
        data_block_desc['data'] = data_block[head_bytes:head_bytes+data_len]
        if cls.data_continued(data_block):
            # Catch data from additional blocks, and append
            data_len_chunks = cls.number_of_chunks(data_len)
            next_block = (head_size + data_len_chunks)*cls.chunk_size
            cont_data_desc = cls.parse_data_block(data_block[next_block:])
            data_block_desc['data'] += cont_data_desc['data']
        return data_block_desc

    @classmethod
    def format_block(cls, block_desc):
        '''Check block type and delegate to correct formatter'''
        if block_desc['block_type'] is 'file':
            return cls.format_file_header(block_desc)
        elif block_desc['block_type'] is 'data':
            return cls.format_data_block(block_desc)
        raise ValueError('Undefined block type:', block_desc['block_type'])

    @classmethod
    def parse_block(cls, block):
        '''Check block type and delegate to correct parser'''
        if cls.header_type(block) is 'file':
            return cls.parse_file_header(block)
        elif cls.header_type(block) is 'data':
            return cls.parse_data_block(block)
        raise ValueError('Undefined header type:', header_bits)



# Define current default formatter
default_formatter = DataFormatter_v1_0



from __future__ import absolute_import
from functools import total_ordering
import sys
from math import fmod

__all__ = ['Timestamp']

if sys.version_info > (3,):
    long = int
    __all__ += ['long']

@total_ordering
class Timestamp(object):
    '''Class for precision time'''
    larpix_clk_freq = long(5e6) #Hz
    larpix_offset_d = long(2**24) #clk cycles

    def __init__(self, ns=0, cpu_time=0, adc_time=0, adj_adc_time=0):
        '''
        Constructor

        Keyword arguments:
        ``ns`` - the serialized timestamp in ns (time since epoch)
        ``cpu_time`` - the serial timestamp used to create ns and adj_adc_time (time since
        epoch)
                - float (expected source is a call to ``time.time()``)
                - must be accurate to at least ``larpix_offset_d/larpix_clk_freq``
        ``adc_time`` - the chip timestamp used to create ``ns`` and ``adj_adc_time`` (clk
        cycles since last rollover)
                - sets the precision of ``Timestamp`` (``1/larpix_clk_freq``)
                - ``int`` between ``0`` and ``larpix_offset_d-1``
        ``adj_adc_time`` - the chip timestamp incremented to account for rollovers (i.e. a
        clk cycle counter)
        '''
        self.ns = long(ns)
        self.cpu_time = float(cpu_time)
        if adc_time // Timestamp.larpix_offset_d != 0:
            raise ValueError('adc_time is out of bounds')
        self.adc_time = int(adc_time)
        self.adj_adc_time = long(adj_adc_time)

    def __str__(self):
        string = 'Timestamp('
        string += 'ns=%d, ' % self.ns
        string += 'cpu_time=%f, ' % self.cpu_time
        string += 'adc_time=%d, ' % self.adc_time
        string += 'adj_adc_time=%d' % self.adj_adc_time
        string += ')'
        return string

    def __repr__(self):
        return str(self)

    def __eq__(self, timestamp):
        '''
        Equality compares only timestamp ns (presumably one only cares about the time
        represented by the Timestamp instance - not how it was calculated).
        '''
        return self.ns == timestamp.ns

    def __gt__(self, timestamp):
        '''
        Ordering is determined only by timestamp ns - see ``__eq__`` note.
        '''
        return self.ns > timestamp.ns

    def __add__(self, timestamp):
        '''
        Sum ns, cpu_time, adc_time, and adj_adc_time of each timestamp.
        Resulting adc_time is mod larpix_offset_d
        '''
        sum_ns = self.ns + timestamp.ns
        sum_cpu = self.cpu_time + timestamp.cpu_time
        sum_adc = int(fmod(self.adc_time + timestamp.adc_time, Timestamp.larpix_offset_d))
        sum_adj_adc = self.adj_adc_time + timestamp.adj_adc_time
        return Timestamp(sum_ns, sum_cpu, sum_adc, sum_adj_adc)

    def __sub__(self, timestamp):
        '''
        Return the difference of ns, cpu_time, adc_time, and adj_adc_time of each timestamp.
        Resulting adc_time is mod larpix_offset_d
        '''
        delta_ns = self.ns - timestamp.ns
        delta_cpu = self.cpu_time - timestamp.cpu_time
        delta_adc = int(fmod(self.adc_time - timestamp.adc_time, Timestamp.larpix_offset_d))
        delta_adj_adc = self.adj_adc_time - timestamp.adj_adc_time
        return Timestamp(delta_ns, delta_cpu, delta_adc, delta_adj_adc)

    def __float__(self):
        '''
        Return a float of sec since epoch
        Note: this is very likely to cause a loss of precision
        '''
        return float(self.ns / 1e9)

    @staticmethod
    def from_packet(packet, cpu_time, ref_time=None):
        '''
        Constructor from packet
        Note: a reference time is required in order to gain precision greater than the cpu
        timestamp, if no reference is provided ns == cpu_time * 1e9 (see description of
        serialized_timestamp() for more details)

        Example usage:
        Generate a list of timestamps from two aligned lists of packets and serial read
        times from a single chip
        ``
        list_of_timestamps = []
        ref_time = from_packet(packets[0], serial_read_time[0])
        for i,packet in enumerate(packets):
            list_of_timestamps += [from_packet(packet, serial_read_time[i], ref_time]
            ref_time = list_of_timestamps[-1]
        ``
        '''
        return Timestamp.serialized_timestamp(adc_time=packet.timestamp, cpu_time=cpu_time,
                                              ref_time=ref_time)

    @staticmethod
    def serialized_timestamp(adc_time, cpu_time, ref_time=None):
        '''
        Generate next serialized packet timestamp from cpu time, adc time, and a reference
        timestamp.

        The algorithm is as follows:
        - If no reference packet, don't add any rollovers (``adj_adc_time == adc_time`` and
        ``ns == cpu_time * 1e9``)
        - If reference packet has the same ``cpu_time``, add rollovers to ``adj_adc_time``
        until ``ref_time.adj_adc_time < adj_adc_time``
        - If reference packet has a different ``cpu_time``, estimate the number of clk cycles
        that have occurred between serial reads add rollovers to ``adj_adc_time`` until the
        difference between ``ref_time.adj_adc_time`` and ``adj_adc_time`` is less than
        ``larpix_offset_d/2`` away from the estimated number of clk cycles
        - The previous two steps provide an ``adj_adc_time`` that is serialized, calculate
        ``ns`` from ``ref_time.ns + <diff btw adj_adc_times>/larpix_clk_freq``

        In order to guarantee that the timestamps are serialized:
        - within a single serial read, the data rate must be greater than
        ``larpix_clk_freq/larpix_offset_d`` (nominally ~.3Hz) OR the serial timeout must be
        less than ``larpix_offset_d/larpix_clk_freq`` (nominally ~3s)
        - between serial reads, the first data packet should arrive within
        ``larpix_offset_d/larpix_clk_freq/2`` of the serial read time (~1.5s)
        - reference time should be the timestamp of the previously read packet from
        that chip (there is a delay between packet creation and packet receipt that can
        interfere with the time ordering of packets between chips)
        - reference time must be a properly sychronized timestamp that was recieved
                - if in the same serial read: within 1 rollover before the current
                timestamp
                - if in separate serial reads: from a previous serial read
        '''
        adj_adc_time = long(adc_time)
        if ref_time is None:
            pass
        elif cpu_time == ref_time.cpu_time:
            # same serial read but adc time may have rolled over
            while adj_adc_time < ref_time.adj_adc_time:
                adj_adc_time += Timestamp.larpix_offset_d
        else:
            # new serial read
            # estimate number of clk cycles between reads
            n_clk_cycles = long((cpu_time - ref_time.ns * 1e-9) * Timestamp.larpix_clk_freq)
            # add offsets until you are close to the predicted number of clk cycles
            while adj_adc_time - ref_time.adj_adc_time - n_clk_cycles \
                    < -Timestamp.larpix_offset_d / 2:
                adj_adc_time += Timestamp.larpix_offset_d

        timestamp = None
        if ref_time is None:
            timestamp = Timestamp(cpu_time * 1e9, cpu_time, adc_time, adj_adc_time)
        else:
            ns = ref_time.ns + long((adj_adc_time - ref_time.adj_adc_time) * long(1e9) \
                                        / Timestamp.larpix_clk_freq)
            timestamp = Timestamp(ns, cpu_time, adc_time, adj_adc_time)
        return timestamp
